# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly status reporter"""
# pylint: disable=protected-access

from itertools import count
from platform import system
from re import match
from unittest.mock import Mock

from pytest import mark, raises

from .status import Status
from .status_reporter import (
    ReductionStatus,
    ReductionStatusReporter,
    StatusReporter,
    TracebackReport,
    _format_duration,
    _format_number,
    _format_seconds,
    main,
)

GBYTES = 1_073_741_824


pytestmark = mark.usefixtures(
    "tmp_path_status_db_fuzz",
    "tmp_path_status_db_reduce",
)


def _fake_sys_info():
    return [
        ("CPU & Load", "64 @ 93% (85.25, 76.21, 51.06)"),
        ("Memory", "183.9GB of 251.9GB free"),
        ("Disk", "22.2GB of 28.7GB free"),
    ]


def test_reduce_status_reporter_01():
    """test basic ReductionStatusReporter"""
    st_rpt = ReductionStatusReporter(None)
    assert not st_rpt.has_results
    st_rpt._sys_info = _fake_sys_info
    assert "No status reports available" in st_rpt.specific()
    assert "No status reports available" in st_rpt.summary()


def test_reduce_status_reporter_02(mocker, tmp_path):
    """test ReductionStatusReporter.load()"""
    # missing reports path
    st_rpt = ReductionStatusReporter.load(tmp_path / "status.db")
    assert not st_rpt.reports

    # empty reports and tb paths
    st_rpt = ReductionStatusReporter.load(tmp_path / "status.db", tb_path=tmp_path)
    assert not st_rpt.reports
    assert isinstance(st_rpt.tracebacks, list)
    assert not st_rpt.tracebacks

    # multiple reports
    size_cb = mocker.Mock(side_effect=count(start=1000, step=-100))
    db_file = tmp_path / "status.db"
    ReductionStatus.start(
        db_file=db_file,
        testcase_size_cb=size_cb,
    )
    status2 = ReductionStatus(
        db_file=db_file,
        pid=1,
        testcase_size_cb=size_cb,
    )
    status2.report(force=True)
    st_rpt = ReductionStatusReporter.load(db_file)
    assert len(st_rpt.reports) > 1


def test_reduce_status_reporter_03(mocker, tmp_path):
    """test ReductionStatusReporter.summary()"""
    mocker.patch("grizzly.common.status.getpid", side_effect=(1, 2))
    mocker.patch("grizzly.common.status.time", side_effect=count(start=1.0, step=1.0))
    size_cb = mocker.Mock(side_effect=count(start=1000, step=-100))
    db_file = tmp_path / "status.db"
    # single report
    status = ReductionStatus.start(
        db_file=db_file,
        testcase_size_cb=size_cb,
        crash_id=123,
        tool="fuzzmatic",
    )
    status.analysis["ran"] = True
    status.run_params["speed"] = 123.0
    status.signature_info["info"] = "crash"
    status.last_reports.append(45678)
    status.record("init")
    with status.measure("total"):
        with status.measure("strategy_0"):
            status.attempts += 1
            status.successes += 1
            status.iterations += 1
        with status.measure("strategy_1"):
            status.attempts += 3
            status.successes += 1
            status.iterations += 3
    status.report(force=True)

    rptr = ReductionStatusReporter.load(db_file)
    rptr._sys_info = _fake_sys_info
    assert rptr.reports
    output = rptr.summary(sysinfo=True, timestamp=True)
    assert "duration" in output
    assert "successes" in output
    assert "attempts" in output
    assert "init" in output
    assert "strategy_0" in output
    assert "strategy_1" in output
    assert "total" in output
    assert "Timestamp" in output
    assert len(output.splitlines()) == 16


def test_reduce_status_reporter_04(mocker, tmp_path):
    """test ReductionStatusReporter.specific()"""
    mocker.patch("grizzly.common.status.getpid", side_effect=(1, 2))
    db_file = tmp_path / "status.db"
    # single report
    status = ReductionStatus.start(
        db_file=db_file,
        strategies=["strategy_0"],
        testcase_size_cb=lambda: 47,
        crash_id=12,
        tool="fuzzmatic",
    )
    assert status.original is None
    rptr = ReductionStatusReporter.load(db_file)
    assert rptr.reports
    output = rptr.specific()
    assert len(output.splitlines()) == 1
    status.analysis["ran"] = True
    status.run_params["splines"] = "reticulated"
    status.last_reports.append(45678)
    status.record("init")
    with status.measure("total"), status.measure("strategy_0"):
        status.iterations = 1
        status.attempts = 1
        status.successes = 1
    status.report(force=True)
    rptr = ReductionStatusReporter.load(db_file)
    assert rptr.reports
    output = rptr.specific()
    assert len(output.splitlines()) == 8
    assert "Analysis" in output
    assert "Run Parameters" in output
    assert "Current Strategy" in output
    assert "Current/Original" in output
    assert "Results" in output
    assert "Time Elapsed" in output


def test_reduce_status_reporter_05(tmp_path):
    """test ReductionStatusReporter.load() with traceback"""
    db_file = tmp_path / "status.db"
    status = ReductionStatus.start(
        db_file=db_file,
        testcase_size_cb=lambda: 47,
    )
    with status.measure("total"):
        status.iteration = 1
    status.report(force=True)
    # create boring screenlog
    (tmp_path / "screenlog.0").write_bytes(b"boring\ntest\n123\n")
    # create first screenlog
    with (tmp_path / "screenlog.1").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"IndexError: list index out of range\n")
    rptr = StatusReporter.load(db_file, tb_path=tmp_path)
    assert len(rptr.tracebacks) == 1
    # create second screenlog
    with (tmp_path / "screenlog.1234").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"foo.bar.error: blah\n")
    rptr = StatusReporter.load(db_file, tb_path=tmp_path)
    assert len(rptr.tracebacks) == 2
    # create third screenlog
    with (tmp_path / "screenlog.3").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"KeyboardInterrupt\n")
    rptr = ReductionStatusReporter.load(db_file, tb_path=tmp_path)
    assert len(rptr.tracebacks) == 2
    merged_log = rptr.summary()
    assert len(merged_log.splitlines()) == 13
    assert "screenlog.1" in merged_log
    assert "screenlog.1234" in merged_log
    assert "IndexError" in merged_log
    assert "foo.bar.error" in merged_log
    assert "screenlog.3" not in merged_log


def test_status_reporter_01():
    """test basic StatusReporter"""
    st_rpt = StatusReporter([])
    assert not st_rpt.has_results
    st_rpt._sys_info = _fake_sys_info
    assert "No status reports available" in st_rpt.specific()
    assert "No status reports available" in st_rpt.summary()


def test_status_reporter_02(tmp_path):
    """test StatusReporter.load()"""
    # missing reports path
    st_rpt = StatusReporter.load(tmp_path / "status.db")
    assert not st_rpt.reports
    # empty reports and tb paths
    st_rpt = StatusReporter.load(tmp_path / "status.db", tb_path=tmp_path)
    assert isinstance(st_rpt.reports, list)
    assert not st_rpt.reports
    assert isinstance(st_rpt.tracebacks, list)
    assert not st_rpt.tracebacks


@mark.parametrize(
    "disk, memory, getloadavg",
    [
        (
            Mock(free=12, total=GBYTES),
            Mock(available=12, total=GBYTES),
            (0.12, 0.34, 0.56),
        ),
        (
            Mock(free=10.23 * GBYTES, total=100 * GBYTES),
            Mock(available=1.1 * GBYTES, total=2 * GBYTES),
            (0.12, 0.34, 0.56),
        ),
        (
            Mock(free=12, total=GBYTES),
            Mock(available=12, total=GBYTES),
            (0.12, 0.34, 0.56),
        ),
    ],
)
def test_status_reporter_03(mocker, disk, memory, getloadavg):
    """test StatusReporter._sys_info()"""
    mocker.patch("grizzly.common.status_reporter.cpu_count", return_value=4)
    mocker.patch("grizzly.common.status_reporter.cpu_percent", return_value=10)
    mocker.patch("grizzly.common.status_reporter.disk_usage", return_value=disk)
    mocker.patch("grizzly.common.status_reporter.virtual_memory", return_value=memory)
    mocker.patch("grizzly.common.status_reporter.getloadavg", return_value=getloadavg)
    sysinfo = StatusReporter._sys_info()
    assert len(sysinfo) == 3
    assert sysinfo[0][0] == "CPU & Load"
    assert sysinfo[1][0] == "Memory"
    assert sysinfo[2][0] == "Disk"
    if system() != "Windows":
        assert sysinfo[0][-1].endswith(" (0.1, 0.3, 0.6)")
    if disk.free < GBYTES or memory.available < GBYTES:
        assert "MB" in sysinfo[1][-1]
        assert "MB" in sysinfo[2][-1]
    else:
        assert "MB" not in sysinfo[1][-1]
        assert "MB" not in sysinfo[2][-1]


def test_status_reporter_04(mocker, tmp_path):
    """test StatusReporter.summary()"""
    mocker.patch("grizzly.common.status.getpid", side_effect=(1, 2))
    mocker.patch("grizzly.common.status.time", side_effect=count(start=1.0, step=1.0))
    db_file = tmp_path / "status.db"
    # single report
    status = Status.start(db_file)
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.report(force=True)
    rptr = StatusReporter.load(db_file)
    rptr._sys_info = _fake_sys_info
    assert rptr.reports is not None
    assert len(rptr.reports) == 1
    output = rptr.summary(runtime=False)
    assert "Iteration" in output
    assert "Instances" not in output
    assert "Rate" in output
    assert "Results" in output
    assert "Blockers" not in output
    assert "Ignored" not in output
    assert "Logs" not in output
    assert "Runtime" not in output
    assert "Timestamp" not in output
    assert len(output.split("\n")) == 3
    # multiple reports
    status = Status.start(db_file)
    status.ignored = 1
    status.iteration = 8
    status.log_size = 86900000
    status.results.count("test", "test")
    status.results.count("test", "test")
    status.report(force=True)
    rptr = StatusReporter.load(db_file)
    rptr._sys_info = _fake_sys_info
    assert len(rptr.reports) == 2
    output = rptr.summary(rate=False, sysinfo=True, timestamp=True)
    assert "Iteration" in output
    assert "Instances" in output
    assert "Rate" not in output
    assert "Results" in output
    assert "Ignored" in output
    assert "Logs" in output
    assert "Runtime" in output
    assert "Timestamp" in output
    lines = output.split("\n")
    assert len(lines) == 10
    # verify alignment
    position = len(lines[0].split(":")[0])
    for line in lines:
        assert match(r"\S\s:\s\S", line[position - 2 :])


def test_status_reporter_05(mocker, tmp_path):
    """test StatusReporter.specific()"""
    mocker.patch("grizzly.common.status.getpid", side_effect=(1, 2))
    db_file = tmp_path / "status.db"
    # single report
    status = Status.start(db_file)
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.report(force=True)
    rptr = StatusReporter.load(db_file)
    assert rptr.reports is not None
    output = rptr.specific()
    assert len(output.strip().split("\n")) == 4
    assert "Ignored" not in output
    assert "Iterations" in output
    assert "Results" in output
    assert "(Blockers detected)" not in output
    assert "Runtime" in output
    # multiple reports
    status = Status.start(db_file, enable_profiling=True)
    status.start_time -= 1220
    status.ignored = 1
    status.iteration = 50
    status.results.count("uid1", "sig1")
    status.results.count("uid1", "sig1")
    status.results.count("uid1", "sig1")
    status.record("test1", 0.91)
    status.record("test1", 1.0)
    status.record("test1", 1.23456)
    status.record("test2", 1201.1)
    status.report(force=True)
    rptr = StatusReporter.load(db_file)
    assert len(rptr.reports) == 2
    output = rptr.specific()
    assert len(output.strip().split("\n")) == 13
    assert "Ignored" in output
    assert "Iterations" in output
    assert "Results" in output
    assert "Runtime" in output
    assert "(Blockers detected)" in output
    assert "Profiling entries" in output
    assert "test1" in output
    assert "test2" in output


def test_status_reporter_06(mocker, tmp_path):
    """test StatusReporter.results()"""
    mocker.patch("grizzly.common.status.getpid", side_effect=(1, 2, 3))
    db_file = tmp_path / "status.db"
    # single report without results
    status = Status.start(db_file)
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.report(force=True)
    rptr = StatusReporter.load(db_file)
    assert rptr.reports is not None
    assert len(rptr.reports) == 1
    assert not rptr.has_results
    assert rptr.results() == "No results available\n"
    # multiple reports with results
    status = Status.start(db_file)
    status.iteration = 1
    status.results.count("uid1", "[@ test1]")
    status.results.count("uid2", "[@ test2]")
    status.results.count("uid1", "[@ test1]")
    status.report(force=True)
    status = Status.start(db_file)
    status.iteration = 1
    status.results.count("uid1", "[@ test1]")
    status.results.count("uid3", "[@ longsignature123]")
    status.report(force=True)
    rptr = StatusReporter.load(db_file)
    assert rptr.has_results
    assert len(rptr.reports) == 3
    output = rptr.results(max_len=19)
    assert "3 : [@ test1]" in output
    assert "1 : [@ test2]" in output
    assert "1 : [@ longsignature..." in output
    assert "(* = Blocker)" in output
    assert len(output.strip().split("\n")) == 4


def test_status_reporter_07(tmp_path):
    """test StatusReporter.load() with traceback"""
    db_file = tmp_path / "status.db"
    status = Status.start(db_file)
    status.ignored = 0
    status.iteration = 1
    status.log_size = 0
    status.report(force=True)
    # create boring screenlog
    (tmp_path / "screenlog.0").write_bytes(b"boring\ntest\n123\n")
    # create first screenlog
    with (tmp_path / "screenlog.1").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"IndexError: list index out of range\n")
    rptr = StatusReporter.load(db_file, tb_path=tmp_path)
    assert len(rptr.tracebacks) == 1
    # create second screenlog
    with (tmp_path / "screenlog.1234").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"foo.bar.error: blah\n")
    rptr = StatusReporter.load(db_file, tb_path=tmp_path)
    assert len(rptr.tracebacks) == 2
    # create third screenlog
    with (tmp_path / "screenlog.3").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"KeyboardInterrupt\n")
    rptr = StatusReporter.load(db_file, tb_path=tmp_path)
    assert len(rptr.tracebacks) == 2
    merged_log = rptr.summary()
    assert len(merged_log.splitlines()) == 14
    assert "screenlog.1" in merged_log
    assert "screenlog.1234" in merged_log
    assert "IndexError" in merged_log
    assert "foo.bar.error" in merged_log
    assert "screenlog.3" not in merged_log


def test_status_reporter_08(tmp_path):
    """test StatusReporter.load() no reports with traceback"""
    # create screenlog with tb
    with (tmp_path / "screenlog.1").open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b"  blah\n")
        test_fp.write(b"IndexError: list index out of range\n")
    rptr = StatusReporter.load(tmp_path / "status.db", tb_path=tmp_path)
    rptr._sys_info = _fake_sys_info
    assert len(rptr.tracebacks) == 1
    output = rptr.summary()
    assert len(output.splitlines()) == 7
    assert "No status reports available" in output
    assert "IndexError" in output


def test_status_reporter_09(mocker, tmp_path):
    """test StatusReporter.summary() limit with traceback"""
    mocker.patch("grizzly.common.status.getpid", side_effect=(1, 2))
    db_file = tmp_path / "status.db"
    # create reports
    status = Status.start(db_file)
    status.ignored = 100
    status.iteration = 1000
    status.log_size = 9999999999
    status.results.count("uid1", "[@ sig1]")
    status.results._count["uid1"] = 123
    status.report(force=True)
    status = Status.start(db_file)
    status.ignored = 9
    status.iteration = 192938
    status.log_size = 0
    status.results.count("uid2", "[@ sig2]")
    status.results._count["uid2"] = 3
    status.report(force=True)
    # create screenlogs with tracebacks
    for i in range(10):
        with (tmp_path / f"screenlog.{i}").open("w") as test_fp:
            test_fp.write("Traceback (most recent call last):\n")
            for j in range(TracebackReport.MAX_LINES):
                test_fp.write(
                    '  File "some/long/path/name/foobar.py", line 5000, in <module>\n'
                )
                test_fp.write(f"    some_long_name_for_a_func_{j:04d}()\n")
            test_fp.write("IndexError: list index out of range\n")
    rptr = StatusReporter.load(db_file, tb_path=tmp_path)
    rptr._sys_info = _fake_sys_info
    assert len(rptr.tracebacks) == 10
    merged_log = rptr.summary(
        runtime=True, sysinfo=True, timestamp=True, iters_per_result=1
    )
    assert len(merged_log) < StatusReporter.SUMMARY_LIMIT


def test_status_reporter_10():
    """test StatusReporter._format_entries()"""
    assert StatusReporter._format_entries([]) == ""
    assert StatusReporter._format_entries([("test", None)]) == "test"
    assert StatusReporter._format_entries([("test", "1")]) == "test : 1"
    out = StatusReporter._format_entries(
        [("first", "1"), ("second", "2"), ("third", "3")]
    )
    assert out == " first : 1\nsecond : 2\n third : 3"


def test_traceback_report_01(tmp_path):
    """test simple TracebackReport"""
    tbr = TracebackReport(
        tmp_path / "log.txt", ["0", "1", "2"], prev_lines=["-2", "-1"]
    )
    output = str(tbr)
    assert len(output.splitlines()) == 6
    assert len(tbr) == 26
    assert "log.txt" in output
    assert "2" in output
    assert "-2" in output


def test_traceback_report_02(tmp_path):
    """test empty TracebackReport"""
    tbr = TracebackReport(tmp_path / "log.txt", [])
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 1
    assert "log.txt" in output


def test_traceback_report_03(tmp_path):
    """test TracebackReport.from_file()"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("w") as test_fp:
        test_fp.write("start junk\npre1\npre2\npre3\npre4\npre5\n")
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 556, in <module>\n')
        test_fp.write("    main(parse_args())\n")
        test_fp.write('  File "foo.py", line 207, in bar\n')
        test_fp.write("    a = b[10]\n")
        test_fp.write("IndexError: list index out of range\n")
        test_fp.write("end junk\n")
    tbr = TracebackReport.from_file(test_log, ignore_kbi=True)
    assert len(tbr.prev_lines) == 5
    assert len(tbr.lines) == 6
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 12
    assert "pre1" in output
    assert "IndexError" in output
    assert "screenlog.0" in output
    assert "junk" not in output

    with test_log.open("w") as test_fp:
        test_fp.write("start junk\n")
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 556, in <module>\n')
        test_fp.write("    main(parse_args())\n")
        test_fp.write('  File "foo.py", line 207, in bar\n')
        test_fp.write("    a = b[10]\n")
        test_fp.write("foo.bar.error: blah\n")
        test_fp.write("end junk\n")
    tbr = TracebackReport.from_file(test_log, max_preceding=0)
    assert len(tbr.lines) == 6
    assert not tbr.prev_lines
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 7
    assert "screenlog.0" in output
    assert "foo.bar.error" in output
    assert "junk" not in output
    # kbi
    with test_log.open("w") as test_fp:
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 556, in <module>\n')
        test_fp.write("    main(parse_args())\n")
        test_fp.write('  File "foo.py", line 207, in bar\n')
        test_fp.write("    a = b[10]\n")
        test_fp.write("KeyboardInterrupt\n")
        test_fp.write("end junk\n")
    tbr = TracebackReport.from_file(test_log)
    assert tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 7
    assert "KeyboardInterrupt" in output
    # ignore kbi
    assert TracebackReport.from_file(test_log, ignore_kbi=True) is None


def test_traceback_report_04(tmp_path):
    """test TracebackReport.from_file() exceed size limit"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("w") as test_fp:
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 5, in <module>\n')
        test_fp.write("    first()\n")
        test_fp.write('  File "foo.py", line 5, in <module>\n')
        test_fp.write("    second()\n")
        for i in reversed(range(TracebackReport.MAX_LINES)):
            test_fp.write('  File "foo.py", line 5, in <module>\n')
            test_fp.write(f"    func_{i:02d}()\n")
        test_fp.write("END_WITH_BLANK_LINE\n\n")
        test_fp.write("end junk\n")
    tbr = TracebackReport.from_file(test_log)
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 18
    assert "<--- TRACEBACK TRIMMED--->" in output
    assert "first()" in output
    assert "func_05()" in output
    assert "second()" not in output
    assert "func_06()" not in output
    assert "END_WITH_BLANK_LINE" in output


def test_traceback_report_05(tmp_path):
    """test TracebackReport.from_file() cut off"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("w") as test_fp:
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 5, in <module>\n')
        test_fp.write("    first()\n")
        for i in range(TracebackReport.MAX_LINES * 2):
            test_fp.write('  File "foo.py", line 5, in <module>\n')
            test_fp.write(f"    func_{i}()\n")
    tbr = TracebackReport.from_file(test_log)
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 18
    assert "first()" in output
    assert f"func_{TracebackReport.MAX_LINES * 2 - 1}" in output


def test_traceback_report_06(tmp_path):
    """test TracebackReport.from_file() single word error"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("w") as test_fp:
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 5, in <module>\n')
        test_fp.write("    first()\n")
        test_fp.write("AssertionError\n")
        test_fp.write("end junk\n")
    tbr = TracebackReport.from_file(test_log)
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 5
    assert "first()" in output
    assert "AssertionError" in output
    assert "end junk" not in output


def test_traceback_report_07(tmp_path):
    """test TracebackReport.from_file() with binary data"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("wb") as test_fp:
        test_fp.write(b"Traceback (most recent call last):\n")
        test_fp.write(b'  File "foo.py", line 5, in <module>\n')
        test_fp.write(b"    bin\xd8()\n")
        test_fp.write(b"AssertionError\n")
    tbr = TracebackReport.from_file(test_log)
    assert not tbr.is_kbi
    output = str(tbr)
    assert "bin()" in output
    assert "AssertionError" in output


def test_traceback_report_08(tmp_path):
    """test TracebackReport.from_file() locate token across chunks"""
    test_log = tmp_path / "screenlog.0"
    with test_log.open("w") as test_fp:
        test_fp.write("A" * (TracebackReport.READ_LIMIT - 5))
        test_fp.write("Traceback (most recent call last):\n")
        test_fp.write('  File "foo.py", line 5, in <module>\n')
        test_fp.write("    first()\n")
        test_fp.write("AssertionError\n")
    tbr = TracebackReport.from_file(test_log)
    assert not tbr.is_kbi
    output = str(tbr)
    assert len(output.splitlines()) == 5
    assert "first()" in output
    assert "AssertionError" in output


@mark.parametrize("report_mode", ["fuzzing", "reducing"])
def test_main_01(mocker, tmp_path, report_mode):
    """test main()"""
    status_db = tmp_path / "status.db"
    # without a report
    assert main([]) == 0
    # with a report
    if report_mode == "fuzzing":
        mocker.patch("grizzly.common.status_reporter.STATUS_DB_FUZZ", status_db)
        status = Status.start(status_db)
        status.results.count("uid", "[@ test]")
    elif report_mode == "reducing":
        mocker.patch("grizzly.common.status_reporter.STATUS_DB_REDUCE", status_db)
        status = ReductionStatus.start(
            status_db,
            testcase_size_cb=lambda: 47,
            strategies=[],
        )
    status.iteration = 1
    status.report(force=True)
    assert main(["--scan-mode", report_mode]) == 0


def test_main_02(capsys):
    """test main() with invalid args"""
    with raises(SystemExit):
        main(["--tracebacks", "missing"])
    assert "--tracebacks must be a directory" in capsys.readouterr()[-1]


@mark.parametrize("report_mode", ["fuzzing", "reducing"])
@mark.parametrize("report_type", ["active", "complete"])
def test_main_03(mocker, tmp_path, report_type, report_mode):
    """test main() --dump"""
    mocker.patch("grizzly.common.status_reporter.cpu_percent", return_value=10)
    status_db = tmp_path / "status.db"
    if report_mode == "reducing":
        mocker.patch("grizzly.common.status_reporter.STATUS_DB_REDUCE", status_db)
        status = ReductionStatus.start(
            status_db,
            testcase_size_cb=lambda: 47,
            strategies=[],
        )
        with status.measure("total"):
            status.iteration = 1
    else:
        mocker.patch("grizzly.common.status_reporter.STATUS_DB_FUZZ", status_db)
        status = Status.start(status_db)
        status.iteration = 1
    status.report(force=True)
    dump_file = tmp_path / "output.txt"
    cmd = ["--dump", str(dump_file), "--type", report_type, "--scan-mode", report_mode]
    assert main(cmd) == 0
    assert dump_file.is_file()
    if report_type == "active":
        assert b"Runtime" not in dump_file.read_bytes()
    else:
        assert b"Timestamp" not in dump_file.read_bytes()


@mark.parametrize(
    "value, expected",
    [
        (0, "0s"),
        (100, "1:40"),
        (3600, "1:00:00"),
    ],
)
def test_format_seconds(value, expected):
    """test _format_seconds used by TableFormatter"""
    assert _format_seconds(value) == expected


@mark.parametrize(
    "value, total, expected",
    [
        (None, 0, ""),
        (0, 0, "0s (  0%)"),
        (100, 0, "1:40 (  0%)"),
        (100, 200, "1:40 ( 50%)"),
        (3600, 3600, "1:00:00 (100%)"),
    ],
)
def test_format_duration(value, total, expected):
    """test _format_duration used by TableFormatter"""
    assert _format_duration(value, total) == expected


@mark.parametrize(
    "value, total, expected",
    [
        (None, 0, ""),
        (0, 0, "0 (  0%)"),
        (100, 0, "100 (  0%)"),
        (100, 200, "100 ( 50%)"),
        (3600, 3600, "3600 (100%)"),
    ],
)
def test_format_number(value, total, expected):
    """test _format_number used by TableFormatter"""
    assert _format_number(value, total) == expected



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Unit tests for `grizzly.reduce.crash` and `grizzly.reduce.bucket`."""
from pytest import mark

from ..common.frontend import Exit
from ..common.reporter import Quality
from .bucket import main as main_wrapper
from .crash import main as crash_main

pytestmark = mark.usefixtures("tmp_path_status_db_reduce")


@mark.parametrize(
    "exit_code, pre_quality, post_quality",
    [
        (Exit.SUCCESS, Quality.UNREDUCED, Quality.ORIGINAL),
        (Exit.ERROR, Quality.UNREDUCED, Quality.REDUCER_ERROR),
        (Exit.ARGS, Quality.REQUEST_SPECIFIC, Quality.UNREDUCED),
        (Exit.ABORT, Quality.REDUCING, Quality.UNREDUCED),
        (Exit.ABORT, Quality.REDUCED, Quality.REDUCED),
        (Exit.LAUNCH_FAILURE, Quality.REQUEST_SPECIFIC, Quality.UNREDUCED),
        (Exit.FAILURE, Quality.UNREDUCED, Quality.NOT_REPRODUCIBLE),
        (Exit.FAILURE, Quality.UNREDUCED, Quality.REQUEST_SPECIFIC),
    ],
)
def test_crash_main_quality(mocker, exit_code, pre_quality, post_quality):
    """test that quality is updated"""
    reduce_main = mocker.patch(
        "grizzly.reduce.crash.ReduceManager.main",
        return_value=exit_code,
    )
    crash = mocker.Mock(testcase_quality=pre_quality, crash_id=1)
    crash.create_signature.side_effect = RuntimeError("no sig to create")
    load_fm_data = mocker.patch("grizzly.reduce.crash.load_fm_data")
    load_fm_data.return_value.__enter__ = mocker.Mock(return_value=(crash, None))
    args = mocker.Mock(
        input=12345,
        no_repro_quality=post_quality.value,
        sig=None,
        tool=None,
    )
    assert crash_main(args) == exit_code
    # verify testcase quality was updated
    assert crash.testcase_quality == post_quality
    assert reduce_main.call_args[0][0].sig is None


def test_bucket_main_wrapper_coverage(mocker):
    """test is for coverage of the wrapper function"""
    mocker.patch("grizzly.reduce.bucket.bucket_main", return_value=0)
    mocker.patch("grizzly.reduce.bucket.ReduceFuzzManagerIDQualityArgs")
    assert main_wrapper() == 0



"""
Sapphire unit tests
"""

# pylint: disable=protected-access

import socket
from random import randint
from threading import Thread, ThreadError

from pytest import mark, raises

from .job import Job
from .worker import Request, Worker


def test_worker_01(mocker):
    """test a Worker"""
    wthread = mocker.Mock(spec_set=Thread)
    wsocket = mocker.Mock(spec_set=socket.socket)
    worker = Worker(wsocket, wthread)
    # it is assumed that launch() has already been called at this point
    assert worker.is_alive()
    assert wthread.is_alive.call_count == 1
    assert not worker.join(timeout=0)
    assert wthread.join.call_count == 1
    # simulator closing a worker that is alive
    # have shutdown raise OSError for coverage
    wsocket.shutdown.side_effect = (OSError("test"),)
    worker.close()
    assert wsocket.shutdown.call_count == 1
    assert wsocket.close.call_count == 1
    # at this point the worker should be complete
    wthread.is_alive.return_value = False
    assert not worker.is_alive()
    assert worker.join(timeout=0)
    # calling a close when the worker is not alive should do nothing
    worker.close()
    assert wsocket.shutdown.call_count == 1
    assert wsocket.close.call_count == 1


@mark.parametrize(
    "exc",
    [
        socket.timeout("test"),
        OSError("test"),
        BlockingIOError("test"),
    ],
)
def test_worker_02(mocker, exc):
    """test Worker.launch() socket exception cases"""
    mocker.patch("sapphire.worker.Thread", autospec=True)
    serv_con = mocker.Mock(spec_set=socket.socket)
    serv_job = mocker.Mock(spec_set=Job)
    serv_con.accept.side_effect = exc
    mocker.patch("sapphire.worker.select", return_value=([serv_con], None, None))
    assert Worker.launch(serv_con, serv_job) is None
    assert serv_job.accepting.clear.call_count == 0
    assert serv_job.accepting.set.call_count == 0


def test_worker_03(mocker):
    """test Worker.launch() thread exception case"""
    mocker.patch("sapphire.worker.sleep", autospec=True)
    mocker.patch("sapphire.worker.Thread", side_effect=ThreadError("test"))
    serv_con = mocker.Mock(spec_set=socket.socket)
    serv_job = mocker.Mock(spec_set=Job)
    conn = mocker.Mock(spec_set=socket.socket)
    serv_con.accept.return_value = (conn, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_con], None, None))
    assert Worker.launch(serv_con, serv_job) is None
    assert conn.close.call_count == 1
    assert serv_job.accepting.clear.call_count == 0
    assert serv_job.accepting.set.call_count == 1


@mark.parametrize(
    "url",
    [
        "/testfile",
        "/./testfile",
        "http://localhost/testfile",
        "http://127.0.0.1/testfile",
        "http://sub.host:1234/testfile",
    ],
)
def test_worker_04(mocker, tmp_path, url):
    """test Worker.launch()"""
    (tmp_path / "testfile").touch()
    job = Job(tmp_path, required_files=["testfile"])
    clnt_sock = mocker.Mock(spec_set=socket.socket)
    clnt_sock.recv.return_value = f"GET {url} HTTP/1.1".encode()
    serv_sock = mocker.Mock(spec_set=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    worker = Worker.launch(serv_sock, job)
    assert worker is not None
    try:
        assert job.is_complete(wait=1)
    finally:
        worker.close()
        if not job.exceptions.empty():
            raise job.exceptions.get()[1]
    assert worker.join(timeout=10)
    assert clnt_sock.sendall.called
    assert serv_sock.accept.call_count == 1
    assert clnt_sock.close.call_count == 1


@mark.parametrize(
    "req, response",
    [
        (b"a", b"400 Bad Request"),
        (b"BAD / HTTP/1.1", b"405 Method Not Allowed"),
    ],
)
def test_worker_05(mocker, tmp_path, req, response):
    """test Worker.launch() with invalid/unsupported requests"""
    (tmp_path / "testfile").touch()
    job = Job(tmp_path, required_files=["testfile"])
    clnt_sock = mocker.Mock(spec_set=socket.socket)
    clnt_sock.recv.return_value = req
    serv_sock = mocker.Mock(spec_set=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    worker = Worker.launch(serv_sock, job)
    assert worker is not None
    assert worker.join(timeout=10)
    worker.close()
    if not job.exceptions.empty():
        raise job.exceptions.get()[1]
    assert serv_sock.accept.call_count == 1
    assert clnt_sock.close.call_count == 1
    assert clnt_sock.sendall.called
    assert response in clnt_sock.sendall.call_args[0][0]


def test_worker_06(mocker):
    """test Worker.handle_request() socket errors"""
    serv_con = mocker.Mock(spec_set=socket.socket)
    serv_con.recv.side_effect = OSError
    serv_job = mocker.Mock(spec_set=Job)
    Worker.handle_request(serv_con, serv_job)
    assert serv_job.accepting.set.call_count == 1
    assert serv_con.sendall.call_count == 0
    assert serv_con.shutdown.call_count == 0
    assert serv_con.close.call_count == 1


def test_response_data_01():
    """test _200_header()"""
    output = Worker._200_header(10, "text/html")
    assert b"Content-Length: 10" in output
    assert b"Content-Type: text/html" in output


def test_response_data_02():
    """test _307_redirect()"""
    output = Worker._307_redirect("http://some.test.url")
    assert b"Location: http://some.test.url" in output


def test_response_data_03():
    """test _4xx_page() without close timeout"""
    output = Worker._4xx_page(400, "Bad Request")
    assert b"Content-Length: " in output
    assert b"HTTP/1.1 400 Bad Request" in output
    assert b"400!" in output


def test_response_data_04():
    """test _4xx_page() with close timeout"""
    output = Worker._4xx_page(404, "Not Found", close=10)
    assert b"Content-Length: " in output
    assert b"HTTP/1.1 404 Not Found" in output
    assert b"window.onload = () => { window.setTimeout(window.close, 10000) }" in output


@mark.parametrize(
    "req, method, scheme, path",
    [
        (b"GET / HTTP/1.1\r\n", "GET", "", "/"),
        (b"GET /foo HTTP/1.1\r\n", "GET", "", "/foo"),
        (b"GET /foo/bar HTTP/1.1\r\n", "GET", "", "/foo/bar"),
        (b"GET http://foo/ HTTP/1.1\r\n", "GET", "http", "/"),
        (b"GET http://foo/bar HTTP/1.1\r\n", "GET", "http", "/bar"),
    ],
)
def test_response_01(req, method, scheme, path):
    """test Request.parse() success"""
    request = Request.parse(req)
    assert request.method == method
    assert request.url.path == path
    assert request.url.scheme == scheme


@mark.parametrize(
    "req",
    [
        b"",
        b"a",
        # Invalid IPv6 URL
        b"GET http://[test/ HTTP/1.1",
        # Missing Path
        b"GET  HTTP/1.1",
        # Invalid Path
        b"GET a a a a a HTTP/1.1",
        # Invalid characters under NFKC normalization
        b"GET http://%E2%84%80/ HTTP/1.1",
        # Missing method
        b" / HTTP/1.1",
    ],
)
def test_response_02(req):
    """test Request.parse() failures"""
    assert Request.parse(req) is None


def test_response_03(mocker):
    """test Request.parse() fail to parse"""
    mocker.patch("sapphire.worker.urlparse", side_effect=ValueError("foo"))
    with raises(ValueError, match="foo"):
        Request.parse(b"GET http://foo HTTP/1.1")


def test_response_04():
    """test Request.parse() by passing random urls"""
    for _ in range(1000):
        # create random 'netloc', for example '%1A%EF%09'
        chars = "".join([f"%{randint(0, 255):02X}" for _ in range(randint(1, 8))])
        Request.parse(f"GET http://{chars}/ HTTP/1.1".encode())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Manage Grizzly status reports."""
from __future__ import annotations

from abc import ABC
from collections import defaultdict
from contextlib import closing, contextmanager
from copy import deepcopy
from dataclasses import astuple, dataclass
from json import dumps, loads
from logging import getLogger
from os import getpid
from sqlite3 import Connection, OperationalError, connect
from time import perf_counter, time
from typing import TYPE_CHECKING, Any, Callable, cast

from .utils import grz_tmp

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from .reporter import FuzzManagerReporter

__all__ = ("ReadOnlyStatus", "ReductionStatus", "SimpleStatus", "Status")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

# time in seconds for db connection to wait before raising an exception
DB_TIMEOUT = 30
# used to track changes to the database layout
DB_VERSION = 2
# database will be updated no more than once within the defined number of seconds.
REPORT_RATE = 60
# default life time for report entries in the database (24 hours)
REPORTS_EXPIRE = 86400
# default life time for result entries in the database (14 days)
RESULTS_EXPIRE = 1209600
# status database files
STATUS_DB_FUZZ = grz_tmp() / "fuzz-status.db"
STATUS_DB_REDUCE = grz_tmp() / "reduce-status.db"

LOG = getLogger(__name__)


@dataclass(eq=False, frozen=True)
class ProfileEntry:
    count: int
    max: float
    min: float
    name: str
    total: float


@dataclass(frozen=True)
class ResultEntry:
    rid: str
    count: int
    desc: str | None


def _db_version_check(con: Connection, expected: int = DB_VERSION) -> bool:
    """Perform version check and remove obsolete tables if required.

    Args:
        con: An open database connection.
        expected: The latest database version.

    Returns:
        True if database was reset otherwise False.
    """
    assert expected > 0
    cur = con.cursor()
    # collect db version and check if an update is required
    cur.execute("PRAGMA user_version;")
    version = cur.fetchone()[0]
    if version < expected:
        cur.execute("BEGIN EXCLUSIVE;")
        # check db version again while locked to avoid race
        cur.execute("PRAGMA user_version;")
        version = cast(int, cur.fetchone()[0])
        if version < expected:
            LOG.debug("db version %d < %d", version, expected)
            # remove ALL tables from the database
            cur.execute("SELECT name FROM sqlite_master WHERE type='table';")
            with con:
                for entry in cur.fetchall():
                    LOG.debug("dropping table %r", entry[0])
                    cur.execute(f"DROP TABLE IF EXISTS {entry[0]};")
                # update db version number
                cur.execute(f"PRAGMA user_version = {expected};")
            return True
    assert version == expected, "code out of date?"
    return False


class SimpleResultCounter:
    __slots__ = ("_count", "_desc", "pid")

    def __init__(self, pid: int) -> None:
        assert pid >= 0
        self._count: dict[str, int] = defaultdict(int)
        self._desc: dict[str, str] = {}
        self.pid = pid

    def __iter__(self) -> Generator[ResultEntry]:
        """Yield all result data.

        Args:
            None

        Yields:
            Contains ID, count and description for each result entry.
        """
        for result_id, count in self._count.items():
            if count > 0:
                yield ResultEntry(result_id, count, self._desc.get(result_id, None))

    def blockers(
        self, iterations: int, iters_per_result: int = 100
    ) -> Generator[ResultEntry]:
        """Any result with an iterations-per-result ratio of less than or equal the
        given limit are considered 'blockers'. Results with a count <= 1 are not
        included.

        Args:
            iterations: Total iterations.
            iters_per_result: Iterations-per-result threshold.

        Yields:
            ResultEntry: ID, count and description of blocking result.
        """
        assert iters_per_result > 0
        if iterations > 0:
            for entry in self:
                if entry.count > 1 and iterations / entry.count <= iters_per_result:
                    yield entry

    def count(self, result_id: str, desc: str) -> tuple[int, bool]:
        """

        Args:
            result_id: Result ID.
            desc: User friendly description.

        Returns:
            Current count for given result_id.
        """
        assert isinstance(result_id, str)
        self._count[result_id] += 1
        initial = False
        if result_id not in self._desc:
            self._desc[result_id] = desc
            initial = True
        return self._count[result_id], initial

    def get(self, result_id: str) -> ResultEntry:
        """Get count and description for given result id.

        Args:
            result_id: Result ID.

        Returns:
            ResultEntry: Count and description.
        """
        assert isinstance(result_id, str)
        return ResultEntry(
            result_id, self._count.get(result_id, 0), self._desc.get(result_id, None)
        )

    @property
    def total(self) -> int:
        """Get total count of all results.

        Args:
            None

        Returns:
            Total result count.
        """
        return sum(self._count.values())


class ReadOnlyResultCounter(SimpleResultCounter):
    def count(self, result_id: str, desc: str) -> tuple[int, bool]:
        raise NotImplementedError("Read only!")  # pragma: no cover

    @classmethod
    def load(cls, db_file: Path, time_limit: float = 0) -> list[ReadOnlyResultCounter]:
        """Load existing entries for database and populate a ReadOnlyResultCounter.

        Args:
            db_file: Database file.
            time_limit: Used to filter older entries.

        Returns:
            Loaded ReadOnlyResultCounter objects.
        """
        assert time_limit >= 0
        with closing(connect(db_file, timeout=DB_TIMEOUT)) as con:
            cur = con.cursor()
            try:
                # collect entries
                if time_limit:
                    cur.execute(
                        """SELECT pid,
                                result_id,
                                description,
                                count
                           FROM results
                           WHERE timestamp > ?;""",
                        (time() - time_limit,),
                    )
                else:
                    cur.execute(
                        """SELECT pid, result_id, description, count FROM results"""
                    )
                entries = cur.fetchall()
            except OperationalError as exc:
                if not str(exc).startswith("no such table:"):
                    raise  # pragma: no cover
                entries = []

        loaded = {}
        for pid, result_id, desc, count in entries:
            if pid not in loaded:
                loaded[pid] = cls(pid)
            loaded[pid]._desc[result_id] = desc  # pylint: disable=protected-access
            loaded[pid]._count[result_id] = count  # pylint: disable=protected-access

        return list(loaded.values())


class ResultCounter(SimpleResultCounter):
    __slots__ = ("_db_file", "_frequent", "_limit", "last_found")

    def __init__(
        self,
        pid: int,
        db_file: Path,
        life_time: int = RESULTS_EXPIRE,
        report_limit: int = 0,
    ) -> None:
        super().__init__(pid)
        assert db_file
        assert report_limit >= 0
        self._db_file = db_file
        self._frequent: set[str] = set()
        # use zero to disable report limit
        self._limit = report_limit
        self.last_found = 0.0
        self._init_db(db_file, pid, life_time)

    @staticmethod
    def _init_db(db_file: Path, pid: int, life_time: float) -> None:
        # prepare database
        LOG.debug("resultcounter using db %s", db_file)
        with closing(connect(db_file, timeout=DB_TIMEOUT)) as con:
            _db_version_check(con)
            cur = con.cursor()
            with con:
                # create table if needed
                cur.execute(
                    """CREATE TABLE IF NOT EXISTS results (
                        count INTEGER NOT NULL,
                        description TEXT NOT NULL,
                        pid INTEGER NOT NULL,
                        result_id TEXT NOT NULL,
                        timestamp INTEGER NOT NULL,
                        PRIMARY KEY(pid, result_id));"""
                )
                # remove expired entries
                if life_time > 0:
                    cur.execute(
                        """DELETE FROM results WHERE timestamp <= ?;""",
                        (time() - life_time,),
                    )
                # avoid (unlikely) pid reuse collision
                cur.execute("""DELETE FROM results WHERE pid = ?;""", (pid,))
                # remove results for jobs that have been removed
                try:
                    cur.execute(
                        """DELETE FROM results
                           WHERE pid NOT IN (SELECT pid FROM status);"""
                    )
                except OperationalError as exc:
                    if not str(exc).startswith("no such table:"):
                        raise  # pragma: no cover

    def count(self, result_id: str, desc: str) -> tuple[int, bool]:
        """Count results and write results to the database.

        Args:
            result_id: Result ID.
            desc: User friendly description.

        Returns:
            Local count and initial report flag (includes parallel instances)
            for given result_id.
        """
        super().count(result_id, desc)
        timestamp = time()
        initial = False
        with closing(connect(self._db_file, timeout=DB_TIMEOUT)) as con:
            cur = con.cursor()
            with con:
                cur.execute(
                    """UPDATE results
                        SET timestamp = ?,
                            count = ?
                        WHERE pid = ?
                        AND result_id = ?;""",
                    (timestamp, self._count[result_id], self.pid, result_id),
                )
                if cur.rowcount < 1:
                    cur.execute(
                        """SELECT pid FROM results WHERE result_id = ?;""",
                        (result_id,),
                    )
                    initial = cur.fetchone() is None
                    cur.execute(
                        """INSERT INTO results(
                                pid,
                                result_id,
                                description,
                                timestamp,
                                count)
                            VALUES (?, ?, ?, ?, ?);""",
                        (self.pid, result_id, desc, timestamp, self._count[result_id]),
                    )
        self.last_found = timestamp
        return self._count[result_id], initial

    def is_frequent(self, result_id: str) -> bool:
        """Scan all results including results from other running instances
        to determine if the limit has been exceeded. Local count must be >1 before
        limit is checked.

        Args:
            result_id: Result ID.

        Returns:
            True if limit has been exceeded otherwise False.
        """
        assert isinstance(result_id, str)
        if self._limit < 1:
            return False
        if result_id in self._frequent:
            return True
        # get local total
        total = self._count.get(result_id, 0)
        # only check the db for parallel results if
        # - result has been found locally more than once
        # - limit has not been exceeded locally
        if self._limit >= total > 1:
            with closing(connect(self._db_file, timeout=DB_TIMEOUT)) as con:
                cur = con.cursor()
                # look up total count from all processes
                cur.execute(
                    """SELECT COALESCE(SUM(count), 0)
                        FROM results WHERE result_id = ?;""",
                    (result_id,),
                )
                global_total = cur.fetchone()[0]
            assert global_total >= total
            total = global_total
        if total > self._limit:
            self._frequent.add(result_id)
            return True
        return False

    def mark_frequent(self, result_id: str) -> None:
        """Mark given results ID as frequent locally.

        Args:
            result_id: Result ID.

        Returns:
            None
        """
        assert isinstance(result_id, str)
        if result_id not in self._frequent:
            self._frequent.add(result_id)


class BaseStatus(ABC):
    """Record and manage status information.

    Attributes:
        _profiles: Profiling data.
        ignored: Ignored result count.
        iteration: Iteration count.
        log_size: Log size in bytes.
        pid: Python process ID.
        start_time: Start time of session.
        test_name: Current test name.
    """

    __slots__ = (
        "_profiles",
        "ignored",
        "iteration",
        "log_size",
        "pid",
        "start_time",
        "test_name",
    )

    def __init__(
        self,
        pid: int,
        start_time: float,
        ignored: int = 0,
        iteration: int = 0,
        log_size: int = 0,
    ) -> None:
        assert pid >= 0
        assert ignored >= 0
        assert iteration >= 0
        assert log_size >= 0
        assert start_time >= 0
        self._profiles: dict[str, dict[str, float | int]] = {}
        self.ignored = ignored
        self.iteration = iteration
        self.log_size = log_size
        self.pid = pid
        self.start_time = start_time
        self.test_name: str | None = None

    def profile_entries(self) -> Generator[ProfileEntry]:
        """Used to retrieve profiling data.

        Args:
            None

        Yields:
            ProfileEntry: Containing recorded profiling data.
        """
        for name, entry in self._profiles.items():
            yield ProfileEntry(
                cast(int, entry["count"]),
                entry["max"],
                entry["min"],
                name,
                entry["total"],
            )

    @property
    def rate(self) -> float:
        """Calculate the average iteration rate in seconds.

        Args:
            None

        Returns:
            Number of iterations performed per second.
        """
        return self.iteration / self.runtime if self.runtime else 0

    @property
    def runtime(self) -> float:
        """Calculate the number of seconds since start() was called.

        Args:
            None

        Returns:
            Total runtime in seconds.
        """
        return max(time() - self.start_time, 0)


class ReadOnlyStatus(BaseStatus):
    """Store status information.

    Attributes:
        _profiles: Profiling data.
        ignored: Ignored result count.
        iteration: Iteration count.
        log_size: Log size in bytes.
        pid: Python process ID.
        results: Result data.
        start_time: Start time of session.
        test_name: Test name.
        timestamp: Last time data was saved to database.
    """

    __slots__ = ("results", "timestamp")

    def __init__(
        self,
        pid: int,
        start_time: float,
        timestamp: float,
        ignored: int = 0,
        iteration: int = 0,
        log_size: int = 0,
        results: ReadOnlyResultCounter | None = None,
    ) -> None:
        super().__init__(
            pid,
            start_time,
            ignored=ignored,
            iteration=iteration,
            log_size=log_size,
        )
        assert timestamp >= start_time
        self.results = results or ReadOnlyResultCounter(pid)
        self.timestamp = timestamp

    @classmethod
    def load_all(
        cls, db_file: Path, time_limit: float = 300
    ) -> Generator[ReadOnlyStatus]:
        """Load all status reports found in `db_file`.

        Args:
            db_file: Database containing status data.
            time_limit: Filter entries by age. Use zero for no limit.

        Yields:
            ReadOnlyStatus: Successfully loaded objects.
        """
        assert time_limit >= 0
        with closing(connect(db_file, timeout=DB_TIMEOUT)) as con:
            cur = con.cursor()
            # collect entries
            try:
                cur.execute(
                    """SELECT pid,
                            _profiles,
                            ignored,
                            iteration,
                            log_size,
                            start_time,
                            timestamp
                        FROM status
                        WHERE timestamp > ?;""",
                    (time() - time_limit if time_limit else 0,),
                )
                entries = cur.fetchall()
            except OperationalError as exc:
                if not str(exc).startswith("no such table:"):
                    raise  # pragma: no cover
                entries = []

        # Load all results
        results = ReadOnlyResultCounter.load(db_file, time_limit=0)
        for entry in entries:
            # look up counter
            current_counter = None
            for counter in results:
                if counter.pid == cast(int, entry[0]):
                    current_counter = counter
                    break

            status = cls(
                entry[0],
                entry[5],
                entry[6],
                ignored=entry[2],
                iteration=entry[3],
                log_size=entry[4],
                results=current_counter,
            )
            status._profiles = loads(entry[1])
            yield status

    @property
    def runtime(self) -> float:
        """Calculate total runtime in seconds relative to 'timestamp'.

        Args:
            None

        Returns:
            Total runtime in seconds.
        """
        return self.timestamp - self.start_time


class SimpleStatus(BaseStatus):
    """Record and manage status information.

    Attributes:
        _profiles: Profiling data.
        ignored: Ignored result count.
        iteration: Iteration count.
        log_size: Log size in bytes.
        pid: Python process ID.
        results:
        start_time: Start time of session.
        test_name: Current test name.
    """

    __slots__ = ("results",)

    def __init__(self, pid: int, start_time: float) -> None:
        super().__init__(pid, start_time)
        self.results = SimpleResultCounter(pid)

    @classmethod
    def start(cls) -> SimpleStatus:
        """Create a unique SimpleStatus object.

        Args:
            None

        Returns:
            Active status report.
        """
        return cls(getpid(), time())


class Status(BaseStatus):
    """Status records status information and stores it in a database.

    Attributes:
        _db_file: Database file containing data.
        _enable_profiling: Profiling support status.
        _profiles: Profiling data.
        ignored: Ignored result count.
        iteration: Iteration count.
        log_size: Log size in bytes.
        pid: Python process ID.
        results: Results data. Used to count occurrences of results.
        start_time: Start time of session.
        test_name: Current test name.
        timestamp: Last time data was saved to database.
    """

    __slots__ = ("_db_file", "_enable_profiling", "results", "timestamp")

    def __init__(
        self,
        pid: int,
        start_time: float,
        db_file: Path,
        enable_profiling: bool = False,
        life_time: float = REPORTS_EXPIRE,
        report_limit: int = 0,
    ) -> None:
        super().__init__(pid, start_time)
        assert life_time >= 0
        assert report_limit >= 0
        self._db_file = db_file
        self._enable_profiling = enable_profiling
        self._init_db(db_file, pid, life_time)
        self.results = ResultCounter(pid, db_file, report_limit=report_limit)
        self.timestamp = start_time

    @staticmethod
    def _init_db(db_file: Path, pid: int, life_time: float) -> None:
        # prepare database
        LOG.debug("status using db %s", db_file)
        with closing(connect(db_file, timeout=DB_TIMEOUT)) as con:
            _db_version_check(con)
            cur = con.cursor()
            with con:
                # create table if needed
                cur.execute(
                    """CREATE TABLE IF NOT EXISTS status (
                        _profiles TEXT NOT NULL,
                        ignored INTEGER NOT NULL,
                        iteration INTEGER NOT NULL,
                        log_size INTEGER NOT NULL,
                        pid INTEGER NOT NULL PRIMARY KEY,
                        start_time REAL NOT NULL,
                        timestamp REAL NOT NULL);"""
                )
                # remove expired status data
                if life_time > 0:
                    cur.execute(
                        """DELETE FROM status WHERE timestamp <= ?;""",
                        (time() - life_time,),
                    )
                # avoid (unlikely) pid reuse collision
                cur.execute("""DELETE FROM status WHERE pid = ?;""", (pid,))

    @contextmanager
    def measure(self, name: str) -> Generator[None]:
        """Used to simplify collecting profiling data.

        Args:
            name: Used to group the entries.

        Yields:
            None
        """
        if self._enable_profiling:
            mark = perf_counter()
            yield
            self.record(name, perf_counter() - mark)
        else:
            yield

    def record(self, name: str, duration: float) -> None:
        """Used to add profiling data. This is intended to be used to make rough
        calculations to identify major configuration issues.

        Args:
            name: Used to group the entries.
            duration: Stored to be later used for measurements.

        Returns:
            None
        """
        if self._enable_profiling:
            assert isinstance(duration, (float, int))
            try:
                self._profiles[name]["count"] += 1
                if self._profiles[name]["max"] < duration:
                    self._profiles[name]["max"] = duration
                elif self._profiles[name]["min"] > duration:
                    self._profiles[name]["min"] = duration
                self._profiles[name]["total"] += duration
            except KeyError:
                # add profile entry
                self._profiles[name] = {
                    "count": 1,
                    "max": duration,
                    "min": duration,
                    "total": duration,
                }

    def report(self, force: bool = False, report_rate: int = REPORT_RATE) -> bool:
        """Write status report to database. Reports are only written periodically.
        It is limited by `report_rate`. The specified number of seconds must
        elapse before another write will be performed unless `force` is True.

        Args:
            force: Ignore report frequency limiting.
            report_rate: Minimum number of seconds between writes to database.

        Returns:
            True if the report was successful otherwise False.
        """
        now = time()
        if self.results.last_found > self.timestamp:
            LOG.debug("results have been found since last report, force update")
            force = True
        assert report_rate >= 0
        if not force and now < (self.timestamp + report_rate):
            return False
        assert self.start_time <= now
        self.timestamp = now

        profiles = dumps(self._profiles)
        with closing(connect(self._db_file, timeout=DB_TIMEOUT)) as con:
            cur = con.cursor()
            with con:
                cur.execute(
                    """UPDATE status
                       SET _profiles = ?,
                           ignored = ?,
                           iteration = ?,
                           log_size = ?,
                           start_time = ?,
                           timestamp = ?
                       WHERE pid = ?;""",
                    (
                        profiles,
                        self.ignored,
                        self.iteration,
                        self.log_size,
                        self.start_time,
                        self.timestamp,
                        self.pid,
                    ),
                )
                if cur.rowcount < 1:
                    cur.execute(
                        """INSERT INTO status(
                               pid,
                               _profiles,
                               ignored,
                               iteration,
                               log_size,
                               start_time,
                               timestamp)
                           VALUES (?, ?, ?, ?, ?, ?, ?);""",
                        (
                            self.pid,
                            profiles,
                            self.ignored,
                            self.iteration,
                            self.log_size,
                            self.start_time,
                            self.timestamp,
                        ),
                    )
        return True

    @classmethod
    def start(
        cls, db_file: Path, enable_profiling: bool = False, report_limit: int = 0
    ) -> Status:
        """Create a unique Status object.

        Args:
            db_file: Database containing status data.
            enable_profiling: Record profiling data.
            report_limit: Number of times a unique result will be reported.

        Returns:
            Status: Active status report.
        """
        status = cls(
            getpid(),
            time(),
            db_file,
            enable_profiling=enable_profiling,
            report_limit=report_limit,
        )
        status.report(force=True)
        return status


@dataclass(frozen=True)
class ReductionStep:
    name: str
    duration: float | None
    successes: int | None
    attempts: int | None
    size: int | None
    iterations: int | None


@dataclass(frozen=True)
class _MilestoneTimer:
    name: str
    start: float
    attempts: int
    iterations: int
    successes: int


class ReductionStatus:
    """Status for a single grizzly reduction"""

    def __init__(
        self,
        strategies: list[str] | None = None,
        testcase_size_cb: Callable[[], int] | None = None,
        crash_id: int | None = None,
        db_file: Path | None = None,
        pid: int | None = None,
        tool: str | None = None,
        life_time: float = REPORTS_EXPIRE,
    ) -> None:
        """Initialize a ReductionStatus instance.

        Arguments:
            strategies: List of strategies to be run.
            testcase_size_cb: Callback to get testcase size.
            crash_id: CrashManager ID of original testcase.
            db_file: Database file containing data.
            tool: The tool name used for reporting to FuzzManager.
            life_time:
        """
        self.analysis: dict[str, float] = {}
        self.attempts = 0
        self.iterations = 0
        # TODO: make RunParams dataclass?
        self.run_params: dict[str, bool | int] = {}
        # TODO: make SigInfo dataclass?
        self.signature_info: dict[str, bool | str] = {}
        self.successes = 0
        self.current_strategy_idx: int | None = None
        self._testcase_size_cb = testcase_size_cb
        self.crash_id = crash_id
        self.finished_steps: list[ReductionStep] = []
        self._in_progress_steps: list[_MilestoneTimer] = []
        self.strategies = strategies
        self._db_file = db_file
        self.pid = pid
        self.timestamp = time()
        self.tool = tool
        self._current_size: int | None = None
        # this holds results from Reporter.submit()
        self.last_reports: list[str] = []

        # prepare database
        if self._db_file:
            LOG.debug("status using db %s", self._db_file)
            with closing(connect(self._db_file, timeout=DB_TIMEOUT)) as con:
                _db_version_check(con)
                cur = con.cursor()
                with con:
                    # create table if needed
                    cur.execute(
                        """CREATE TABLE IF NOT EXISTS reduce_status (
                           pid INTEGER NOT NULL PRIMARY KEY,
                           analysis TEXT NOT NULL,
                           attempts INTEGER NOT NULL,
                           iterations INTEGER NOT NULL,
                           run_params TEXT NOT NULL,
                           signature_info TEXT NOT NULL,
                           successes INTEGER NOT NULL,
                           crash_id INTEGER,
                           finished_steps TEXT NOT NULL,
                           _in_progress_steps TEXT NOT NULL,
                           strategies TEXT NOT NULL,
                           _current_size INTEGER NOT NULL,
                           current_strategy_idx INTEGER,
                           timestamp REAL NOT NULL,
                           tool TEXT,
                           last_reports TEXT NOT NULL);"""
                    )
                    # remove expired status data
                    if life_time > 0:
                        cur.execute(
                            """DELETE FROM reduce_status WHERE timestamp <= ?;""",
                            (time() - life_time,),
                        )
                    # avoid (unlikely) pid reuse collision
                    cur.execute("""DELETE FROM reduce_status WHERE pid = ?;""", (pid,))

    @classmethod
    def start(
        cls,
        db_file: Path,
        strategies: list[str] | None = None,
        testcase_size_cb: Callable[[], int] | None = None,
        crash_id: int | None = None,
        tool: str | None = None,
    ) -> ReductionStatus:
        """Create a unique ReductionStatus object.

        Args:
            db_file: Database containing status data.
            strategies: List of strategies to be run.
            testcase_size_cb: Callback to get testcase size.
            crash_id: CrashManager ID of original testcase.
            tool: The tool name used for reporting to FuzzManager.

        Returns:
            Active status report.
        """
        status = cls(
            crash_id=crash_id,
            db_file=db_file,
            pid=getpid(),
            strategies=strategies,
            testcase_size_cb=testcase_size_cb,
            tool=tool,
        )
        status.report(force=True)
        return status

    def report(self, force: bool = False, report_rate: float = REPORT_RATE) -> bool:
        """Write status report to database. Reports are only written periodically.
        It is limited by `report_rate`. The specified number of seconds must
        elapse before another write will be performed unless `force` is True.

        Args:
            force: Ignore report frequently limiting.
            report_rate: Minimum number of seconds between writes.

        Returns:
            True if the report was successful otherwise false.
        """
        now = time()
        if not force and now < (self.timestamp + report_rate):
            return False
        assert self._db_file
        self.timestamp = now

        with closing(connect(self._db_file, timeout=DB_TIMEOUT)) as con:
            cur = con.cursor()
            with con:
                analysis = dumps(self.analysis)
                run_params = dumps(self.run_params)
                sig_info = dumps(self.signature_info)
                finished = dumps([astuple(step) for step in self.finished_steps])
                in_prog = dumps([astuple(step) for step in self._in_progress_steps])
                strategies = dumps(self.strategies)
                last_reports = dumps(self.last_reports)

                cur.execute(
                    """UPDATE reduce_status
                       SET analysis = ?,
                           attempts = ?,
                           iterations = ?,
                           run_params = ?,
                           signature_info = ?,
                           successes = ?,
                           crash_id = ?,
                           finished_steps = ?,
                           _in_progress_steps = ?,
                           strategies = ?,
                           _current_size = ?,
                           current_strategy_idx = ?,
                           timestamp = ?,
                           tool = ?,
                           last_reports = ?
                       WHERE pid = ?;""",
                    (
                        analysis,
                        self.attempts,
                        self.iterations,
                        run_params,
                        sig_info,
                        self.successes,
                        self.crash_id,
                        finished,
                        in_prog,
                        strategies,
                        self._testcase_size(),
                        self.current_strategy_idx,
                        self.timestamp,
                        self.tool,
                        last_reports,
                        self.pid,
                    ),
                )
                if cur.rowcount < 1:
                    cur.execute(
                        """INSERT INTO reduce_status(
                               pid,
                               analysis,
                               attempts,
                               iterations,
                               run_params,
                               signature_info,
                               successes,
                               crash_id,
                               finished_steps,
                               _in_progress_steps,
                               strategies,
                               _current_size,
                               current_strategy_idx,
                               timestamp,
                               tool,
                               last_reports)
                           VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);""",
                        (
                            self.pid,
                            analysis,
                            self.attempts,
                            self.iterations,
                            run_params,
                            sig_info,
                            self.successes,
                            self.crash_id,
                            finished,
                            in_prog,
                            strategies,
                            self._testcase_size(),
                            self.current_strategy_idx,
                            self.timestamp,
                            self.tool,
                            last_reports,
                        ),
                    )

        return True

    @classmethod
    def load_all(
        cls, db_file: Path, time_limit: float = 300
    ) -> Generator[ReductionStatus]:
        """Load all reduction status reports found in `db_file`.

        Args:
            db_file: Database containing status data.
            time_limit: Only include entries with a timestamp that is within the
                        given number of seconds. Use zero for no limit.

        Yields:
            Successfully loaded read-only status objects.
        """
        assert time_limit >= 0
        with closing(connect(db_file, timeout=DB_TIMEOUT)) as con:
            cur = con.cursor()
            # collect entries
            try:
                cur.execute(
                    """SELECT pid,
                              analysis,
                              attempts,
                              iterations,
                              run_params,
                              signature_info,
                              successes,
                              crash_id,
                              finished_steps,
                              _in_progress_steps,
                              strategies,
                              _current_size,
                              current_strategy_idx,
                              timestamp,
                              tool,
                              last_reports
                       FROM reduce_status
                       WHERE timestamp > ?
                       ORDER BY timestamp DESC;""",
                    (time() - time_limit if time_limit else 0,),
                )
                entries = cur.fetchall()
            except OperationalError as exc:
                if not str(exc).startswith("no such table:"):
                    raise  # pragma: no cover
                entries = []

        for entry in entries:
            pid = entry[0]

            status = cls(
                strategies=loads(entry[10]),
                crash_id=entry[7],
                pid=pid,
                tool=entry[14],
            )
            status.analysis = loads(entry[1])
            status.attempts = entry[2]
            status.iterations = entry[3]
            status.run_params = loads(entry[4])
            status.signature_info = loads(entry[5])
            status.successes = entry[6]
            status.finished_steps = [ReductionStep(*step) for step in loads(entry[8])]
            status._in_progress_steps = [
                _MilestoneTimer(*step) for step in loads(entry[9])
            ]
            status._current_size = entry[11]
            status.current_strategy_idx = entry[12]
            status.timestamp = entry[13]
            status.last_reports = loads(entry[15])
            yield status

    def _testcase_size(self) -> int | None:
        if self._db_file is None:
            return self._current_size
        assert self._testcase_size_cb is not None
        return self._testcase_size_cb()

    def __deepcopy__(self, memo: dict[int, Any] | None) -> ReductionStatus:
        """Return a deep copy of this instance."""
        # pylint: disable=protected-access
        result = type(self)(
            strategies=deepcopy(self.strategies, memo),
            crash_id=self.crash_id,
            testcase_size_cb=self._testcase_size_cb,
            pid=self.pid,
            tool=self.tool,
        )
        # assign after construction to avoid DB access
        result._db_file = self._db_file
        result.analysis = deepcopy(self.analysis, memo)
        result.attempts = self.attempts
        result.iterations = self.iterations
        result.run_params = deepcopy(self.run_params, memo)
        result.signature_info = deepcopy(self.signature_info, memo)
        result.successes = self.successes
        result.finished_steps = deepcopy(self.finished_steps, memo)
        result.last_reports = deepcopy(self.last_reports, memo)
        # finish open timers
        for tmr in reversed(self._in_progress_steps):
            step = self._tmr_to_step(tmr)
            result.record(
                step.name,
                attempts=step.attempts,
                duration=step.duration,
                iterations=step.iterations,
                successes=step.successes,
                report=False,
            )
        return result

    def _tmr_to_step(self, tmr: _MilestoneTimer) -> ReductionStep:
        if self._db_file is None:
            duration = self.timestamp - tmr.start
        else:
            duration = time() - tmr.start
        return ReductionStep(
            tmr.name,
            duration=duration,
            successes=self.successes - tmr.successes,
            attempts=self.attempts - tmr.attempts,
            size=None,
            iterations=self.iterations - tmr.iterations,
        )

    @property
    def current_strategy(self) -> ReductionStep | None:
        if self._in_progress_steps:
            return self._tmr_to_step(self._in_progress_steps[-1])
        if self.finished_steps:
            return self.finished_steps[-1]
        return None

    @property
    def total(self) -> ReductionStep | None:
        if self._in_progress_steps:
            return self._tmr_to_step(self._in_progress_steps[0])
        if self.finished_steps:
            return self.finished_steps[-1]
        return None

    @property
    def original(self) -> ReductionStep | None:
        if self.finished_steps:
            return self.finished_steps[0]
        return None

    def record(
        self,
        name: str,
        duration: float | None = None,
        iterations: int | None = None,
        attempts: int | None = None,
        successes: int | None = None,
        report: bool = True,
    ) -> None:
        """Record reduction status for a given point in time:

        - name of the milestone (eg. init, strategy name completed)
        - elapsed time (seconds)
        - # of iterations
        - # of total attempts
        - # of successful attempts

        Arguments:
            name: name of milestone
            duration: seconds elapsed for period recorded
            iterations: # of iterations performed
            attempts: # of attempts performed
            successes: # of attempts successful
            report: Automatically force a report.

        Returns:
            None
        """
        self.finished_steps.append(
            ReductionStep(
                name=name,
                size=self._testcase_size(),
                duration=duration,
                iterations=iterations,
                attempts=attempts,
                successes=successes,
            )
        )
        if report:
            self.report(force=True)

    @contextmanager
    def measure(self, name: str, report: bool = True) -> Generator[None]:
        """Time and record the period leading up to a reduction milestone.
        eg. a strategy being run.

        Arguments:
            name: name of milestone
            report: Automatically force a report.

        Yields:
            None
        """

        tmr = _MilestoneTimer(
            name, time(), self.attempts, self.iterations, self.successes
        )
        self._in_progress_steps.append(tmr)
        yield
        assert self._in_progress_steps.pop() is tmr
        step = self._tmr_to_step(tmr)
        self.record(
            name,
            attempts=step.attempts,
            duration=step.duration,
            iterations=step.iterations,
            successes=step.successes,
            report=report,
        )

    def copy(self) -> ReductionStatus:
        """Create a deep copy of this instance.

        Arguments:
            None

        Returns:
            Clone of self
        """
        return deepcopy(self)

    def add_to_reporter(
        self, reporter: FuzzManagerReporter, expected: bool = True
    ) -> None:
        """Add the reducer status to reported metadata for the given reporter.

        Arguments:
            reporter: Reporter to update.
            expected: Add detailed stats.

        Returns:
            None
        """
        # only add detailed stats for expected results
        if expected:
            reporter.add_extra_metadata(
                "reducer-stats", [astuple(step) for step in self.finished_steps]
            )
        # other parameters
        if self.analysis:
            reporter.add_extra_metadata("reducer-analysis", self.analysis)
        if self.run_params:
            reporter.add_extra_metadata("reducer-params", self.run_params)
        if self.signature_info:
            reporter.add_extra_metadata("reducer-sig", self.signature_info)
        # if input was an existing crash-id, record the original
        if self.crash_id:
            reporter.add_extra_metadata("reducer-input", self.crash_id)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Unit tests for `grizzly.reduce.main`."""
from pytest import mark, raises

from ..common.frontend import ConfigError, Exit
from ..common.storage import TestCase, TestCaseLoadFailure
from ..target import AssetManager, Target, TargetLaunchError, TargetLaunchTimeout
from . import ReduceManager
from .args import ReduceArgs, ReduceFuzzManagerIDArgs, ReduceFuzzManagerIDQualityArgs
from .exceptions import GrizzlyReduceBaseException

pytestmark = mark.usefixtures("tmp_path_status_db_reduce")


def test_args_01(capsys, tmp_path, mocker):
    """test args in common with grizzly.replay"""
    # pylint: disable=import-outside-toplevel
    from ..replay.test_args import test_replay_args_01 as real_test

    mocker.patch("grizzly.replay.test_args.ReplayArgs", new=ReduceArgs)
    real_test(capsys, mocker, tmp_path)


def test_args_02(capsys, tmp_path):
    """test parsing args specific to grizzly.reduce"""
    exe = tmp_path / "binary"
    exe.touch()
    inp = tmp_path / "input"
    inp.mkdir()
    (inp / "somefile").touch()

    # test valid strategy
    ReduceArgs().parse_args([str(exe), str(inp), "--strategy", "lines"])
    # test invalid strategy
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--strategy", "cosmic_radiation"])
    assert (
        "error: argument --strategy: invalid choice: 'cosmic_radiation'"
        in capsys.readouterr()[-1]
    )
    # test no-analysis
    ReduceArgs().parse_args(
        [str(exe), str(inp), "--no-analysis", "--repeat", "99", "--min-crashes", "99"]
    )
    # test multiple inputs
    ReduceArgs().parse_args([str(exe), str(inp), str(inp)])
    # test no-harness
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), str(inp), "--no-harness"])
    assert (
        "error: '--no-harness' cannot be used with multiple testcases"
        in capsys.readouterr()[-1]
    )
    # these should both log a warning that the args will be ignored due to analysis
    ReduceArgs().parse_args([str(exe), str(inp), "--repeat", "99"])
    ReduceArgs().parse_args([str(exe), str(inp), "--min-crashes", "99"])
    ReduceArgs().parse_args([str(exe), str(inp), "--report-period", "99"])
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--report-period", "0"])
    assert "error: Invalid --report-period" in capsys.readouterr()[-1]
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--report-period", "15"])
    assert "error: Very short --report-period" in capsys.readouterr()[-1]


def test_args_03(tmp_path, capsys):
    """test ReduceFuzzManagerIDArgs"""
    exe = tmp_path / "binary"
    exe.touch()
    ReduceFuzzManagerIDArgs().parse_args([str(exe), "123"])
    # error cases
    with raises(SystemExit):
        ReduceFuzzManagerIDArgs().parse_args(
            [str(exe), "123", "--no-harness", "--test-index", "0", "1"]
        )
    assert (
        "error: '--test-index' only supports a single value with '--no-harness'"
        in capsys.readouterr()[-1]
    )


def test_args_04(tmp_path):
    """test ReduceFuzzManagerIDQualityArgs"""
    exe = tmp_path / "binary"
    exe.touch()
    ReduceFuzzManagerIDQualityArgs().parse_args([str(exe), "123"])


@mark.parametrize(
    "patch_func, side_effect, return_value, kwargs, exit_code, use_sig",
    [
        (
            "grizzly.reduce.core.ReduceManager.run",
            TargetLaunchError("error", None),
            None,
            {},
            Exit.LAUNCH_FAILURE,
            False,
        ),
        (
            "grizzly.reduce.core.ReduceManager.run",
            TargetLaunchTimeout,
            None,
            {},
            Exit.LAUNCH_FAILURE,
            False,
        ),
        (
            "grizzly.reduce.core.load_plugin",
            KeyboardInterrupt,
            None,
            {},
            Exit.ABORT,
            True,
        ),
        (
            "grizzly.reduce.core.load_plugin",
            GrizzlyReduceBaseException("", 999),
            None,
            {},
            999,
            False,
        ),
        (
            "grizzly.reduce.core.ReplayManager.load_testcases",
            TestCaseLoadFailure,
            None,
            {},
            Exit.ERROR,
            False,
        ),
        (
            "grizzly.reduce.core.ReplayManager.load_testcases",
            ConfigError("", 999),
            None,
            {"no_harness": False},
            999,
            False,
        ),
        (
            "grizzly.reduce.core.ReplayManager.load_testcases",
            RuntimeError,
            None,
            {},
            Exit.ERROR,
            False,
        ),
    ],
)
def test_main_exit(
    mocker, tmp_path, patch_func, side_effect, return_value, kwargs, exit_code, use_sig
):
    """test ReduceManager.main() failure cases"""
    mocker.patch("grizzly.reduce.core.FuzzManagerReporter", autospec=True)
    mocker.patch("grizzly.reduce.core.load_plugin", autospec=True)
    mocker.patch("grizzly.reduce.core.ReductionStatus", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)

    if use_sig:
        sig = tmp_path / "fake.sig"
        sig.write_text('{"symptoms": [{"address": "0", "type": "crashAddress"}]}')
        sig.with_suffix(".metadata").write_text('{"shortDescription": "foo"}')
    else:
        sig = None

    (tmp_path / "test.html").touch()
    # setup args
    args = mocker.Mock(
        ignore=["fake"],
        input=[tmp_path / "test.html"],
        min_crashes=1,
        relaunch=1,
        repeat=1,
        sig=sig,
        test_index=[],
        timeout=10,
        **kwargs
    )

    mocker.patch(patch_func, side_effect=side_effect, return_value=return_value)
    assert ReduceManager.main(args) == exit_code


@mark.parametrize(
    "exc_type",
    [
        TargetLaunchError,
        TargetLaunchTimeout,
    ],
)
def test_main_launch_error(mocker, exc_type):
    mocker.patch("grizzly.reduce.core.ReductionStatus", autospec=True)
    mocker.patch("grizzly.reduce.core.FuzzManagerReporter", autospec=True)
    reporter = mocker.patch("grizzly.reduce.core.FailedLaunchReporter", autospec=True)
    mocker.patch("grizzly.reduce.core.load_plugin", autospec=True)
    mocker.patch(
        "grizzly.reduce.core.ReplayManager.load_testcases",
        return_value=(
            [mocker.Mock(spec_set=TestCase, hang=False, adapter_name="fake")],
            None,
            {},
        ),
    )
    mocker.patch("grizzly.reduce.core.time_limits", return_value=(None, 10))
    run = mocker.patch("grizzly.reduce.core.ReduceManager.run", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)
    # setup args
    args = mocker.Mock(
        ignore=["fake"],
        input=["test"],
        min_crashes=1,
        no_harness=False,
        relaunch=1,
        repeat=1,
        sig=None,
        tool=None,
    )

    exc_obj = exc_type(
        "error", mocker.Mock() if exc_type is TargetLaunchError else None
    )
    run.side_effect = exc_obj
    assert ReduceManager.main(args) == 4
    if exc_type is TargetLaunchError:
        assert reporter.return_value.submit.call_count == 1
        reported_testcases, report = reporter.return_value.submit.call_args[0]
        assert reported_testcases == []
        assert report is exc_obj.report
    else:
        assert reporter.return_value.submit.call_count == 0


@mark.parametrize("https_supported", [True, False])
def test_main_https_support(mocker, tmp_path, https_supported):
    """test ReduceManager.main() - Target HTTPS support"""
    mocker.patch("grizzly.reduce.core.FuzzManagerReporter", autospec=True)
    mocker.patch("grizzly.reduce.core.get_certs", autospec=True)
    mocker.patch("grizzly.reduce.core.ReduceManager.run", autospec=True, return_value=0)
    mocker.patch("grizzly.reduce.core.ReductionStatus", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)
    (tmp_path / "test.html").touch()
    # setup args
    args = mocker.Mock(
        ignore=["fake"],
        input=[tmp_path / "test.html"],
        min_crashes=1,
        relaunch=1,
        repeat=1,
        sig=None,
        test_index=[],
        use_http=False,
        time_limit=1,
        timeout=1,
    )

    target_cls = mocker.MagicMock(spec_set=Target)
    target = target_cls.return_value
    target.https.return_value = https_supported
    mocker.patch("grizzly.reduce.core.load_plugin", return_value=target_cls)
    assert ReduceManager.main(args) == 0
    assert target.https.call_count == 1


def test_main_load_assets_and_env(mocker, tmp_path):
    """test ReduceManager.main() - Use assets and env vars from loaded TestCase"""
    asset_mgr = mocker.Mock(spec_set=AssetManager)
    mocker.patch(
        "grizzly.reduce.core.ReplayManager.load_testcases",
        autospec=True,
        return_value=([], asset_mgr, {"FOO_ENV": "123"}),
    )
    mocker.patch("grizzly.reduce.core.ReduceManager.run", autospec=True, return_value=0)
    mocker.patch("grizzly.reduce.core.ReductionStatus", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)
    (tmp_path / "test.html").touch()
    # setup args
    args = mocker.Mock(
        ignore=["fake"],
        input=[tmp_path / "test.html"],
        min_crashes=1,
        relaunch=1,
        repeat=1,
        sig=None,
        test_index=[],
        time_limit=1,
        timeout=1,
    )

    target = mocker.Mock(spec_set=Target)
    target_cls = mocker.MagicMock(spec_set=Target, return_value=target)
    mocker.patch("grizzly.reduce.core.load_plugin", return_value=target_cls)
    assert ReduceManager.main(args) == 0
    assert target.merge_environment.call_count == 1
    assert target.merge_environment.call_args.args == ({"FOO_ENV": "123"},)
    assert asset_mgr.add_batch.call_count == 1
    # this should not be called since the AssetManager was given to the target
    # and the target is a mock
    assert asset_mgr.cleanup.call_count == 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pytest import raises

from .cache import _active_cache, add_cached, clear_cached, find_cached


def test_cache_active_cache(mocker, tmp_path):
    """test _active_cache()"""
    cache_path = tmp_path / "cache"
    cache_path.mkdir()
    mocker.patch("grizzly.common.cache.CACHE_PATH", new=cache_path)
    mocker.patch("grizzly.common.cache.LOCK_FILE", new=tmp_path / "cache.lock")
    # create path
    mocker.patch("grizzly.common.cache._ACTIVE_CACHE", new=None)
    new_path = _active_cache()
    assert any(cache_path.iterdir())
    assert cache_path in new_path.parents
    # find existing with expired entry
    mocker.patch("grizzly.common.cache._ACTIVE_CACHE", new=None)
    (cache_path / "0").mkdir()
    found_path = _active_cache()
    assert found_path == new_path


def test_cache_basic(mocker, tmp_path):
    """test basic cache functionality"""
    mocker.patch("grizzly.common.cache._ACTIVE_CACHE", new=None)
    (tmp_path / "cache").mkdir()
    mocker.patch("grizzly.common.cache.CACHE_PATH", new=tmp_path / "cache")
    mocker.patch("grizzly.common.cache.LOCK_FILE", new=tmp_path / "cache.lock")
    # attempt to clear empty cache
    clear_cached()
    # invalid keys
    with raises(ValueError, match="Key must be alphanumeric"):
        add_cached("", tmp_path / "foo")
    with raises(ValueError, match="Key must be alphanumeric"):
        find_cached("test/key")
    # look for non-existing entry
    assert find_cached("test-key") is None
    # add entry
    content = tmp_path / "content"
    content.mkdir()
    (content / "data.txt").write_text("123")
    cache_dir = add_cached("test-key", content)
    assert cache_dir.is_dir()
    assert not content.is_dir()
    assert (cache_dir / "content" / "data.txt").is_file()
    assert (cache_dir / "content" / "data.txt").read_text() == "123"
    # find existing entry
    cache_dir = find_cached("test-key")
    assert cache_dir is not None
    assert (cache_dir / "content" / "data.txt").is_file()
    assert (cache_dir / "content" / "data.txt").read_text() == "123"
    # look for missing entry when cache has data
    assert find_cached("missing") is None
    # add when entry exists (use existing)
    collision = add_cached("test-key", content)
    assert collision is not None
    assert (collision / "content" / "data.txt").samefile(
        cache_dir / "content" / "data.txt"
    )
    # clear with nothing expired
    clear_cached()
    assert find_cached("test-key") is not None
    # clear with everything expired
    clear_cached(max_age=0)
    assert find_cached("test-key") is None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import binascii
from base64 import b64decode
from logging import getLogger
from os import environ
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from types import MappingProxyType
from typing import TYPE_CHECKING
from zipfile import ZipFile

from bugsy import Bug, Bugsy
from bugsy.errors import BugsyException
from requests.exceptions import ConnectionError as RequestsConnectionError

from .utils import grz_tmp

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

# attachments that can be ignored
IGNORE_EXTS = frozenset(
    (
        "c",
        "cpp",
        "diff",
        "diffs",
        "exe",
        "gz",
        "idl",
        "log",
        "patch",
        "php",
        "py",
        "tar",
        "txt",
        "xul",
    )
)
# TODO: support all target assets
KNOWN_ASSETS = MappingProxyType({"prefs": "prefs.js"})
LOG = getLogger(__name__)


class BugzillaBug:
    __slots__ = ("_bug", "_data")

    def __init__(self, bug: Bug, ignore: Iterable[str] = IGNORE_EXTS) -> None:
        self._bug = bug
        self._data = Path(
            mkdtemp(prefix=f"bug{bug.id}-", dir=grz_tmp("bugzilla"))
        ).resolve()
        self._fetch_attachments(ignore)

    def __enter__(self) -> BugzillaBug:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    def _fetch_attachments(
        self, ignore: Iterable[str], require_extension: bool = True
    ) -> None:
        """Download bug attachments.

        Arguments:
            ignore: File extensions to ignore.
            require_extension: Attachment must have a file extension.

        Returns:
            None
        """
        for attachment in self._bug.get_attachments():
            # pylint: disable=too-many-boolean-expressions
            if (
                attachment.is_obsolete
                or attachment.content_type == "text/x-phabricator-request"
                or not attachment.file_name
                or (require_extension and "." not in attachment.file_name)
                or attachment.file_name.split(".")[-1] in ignore
            ):
                continue
            target = (self._data / attachment.file_name).resolve()
            # check file name
            if target.is_dir() or not target.is_relative_to(self._data):
                LOG.debug(
                    "bug %d: skipping attachment %r", self._bug.id, attachment.file_name
                )
                continue
            # decode data
            try:
                data = b64decode(attachment.data or "")
            except binascii.Error as exc:
                LOG.warning(
                    "Failed to decode attachment: %r (%s)", attachment.file_name, exc
                )
                continue
            target.write_bytes(data)

    def _unpack_archives(self) -> None:
        """Unpack and remove archives.

        Arguments:
            None

        Returns:
            None
        """
        for num, entry in enumerate(self._data.iterdir()):
            if entry.suffix.lower() == ".zip" and entry.is_file():
                dst = self._data / f"unpacked_{num:02d}_{entry.stem}"
                LOG.debug("unpacking %s to '%s'", entry, dst)
                with ZipFile(entry) as zip_fp:
                    zip_fp.extractall(path=dst)
                entry.unlink()
            # TODO: add support for other archive types

    def assets(self, ignore: tuple[str] | None = None) -> Generator[tuple[str, Path]]:
        """Scan files for assets.

        Arguments:
            ignore: Assets not to include in output.

        Yields:
            Asset name and path.
        """
        for asset, file in KNOWN_ASSETS.items():
            if not ignore or asset not in ignore:
                asset_path = self._data / file
                if asset_path.is_file():
                    yield asset, asset_path

    def cleanup(self) -> None:
        """Remove attachment data.

        Arguments:
            None

        Returns:
            None
        """
        if self._data.exists():
            rmtree(self._data)

    @classmethod
    def load(cls, bug_id: int) -> BugzillaBug | None:
        """Load bug information from a Bugzilla instance.

        Arguments:
            bug_id: Bug to load.

        Returns:
            BugzillaBug
        """
        api_key = environ.get("BZ_API_KEY")
        # default root matches Bugsy
        api_root = environ.get("BZ_API_ROOT", "https://bugzilla.mozilla.org/rest")
        bugzilla = Bugsy(api_key=api_key, bugzilla_url=api_root)
        try:
            return cls(bugzilla.get(bug_id))
        except BugsyException as exc:
            LOG.error("%s", exc.msg)
            # Access Denied
            if api_key is None and exc.code == 102:
                LOG.warning(
                    "Set BZ_API_KEY in your environment or download the testcase "
                    "manually to run with grizzly.replay."
                )
        except RequestsConnectionError as exc:
            LOG.error("Unable to connect to %r (%s)", bugzilla.bugzilla_url, exc)
        return None

    @property
    def path(self) -> Path:
        """Directory containing downloaded bug attachments.

        Arguments:
            None

        Returns:
            Path to directory.
        """
        return self._data

    def testcases(self) -> list[Path]:
        """Create a list of potential test cases.

        Arguments:
            None

        Returns:
            Files and directories that could potentially be test cases.
        """
        # unpack archives
        self._unpack_archives()
        testcases = [x for x in self._data.iterdir() if x.is_dir()]
        # scan base directory for files, filtering out assets
        files = tuple(
            x
            for x in self._data.iterdir()
            if x.is_file() and x.name.lower() not in KNOWN_ASSETS.values()
        )
        # first, if base directory contains multiple files add it as a single test case
        if len(files) > 1:
            testcases.append(self._data)
        # finally, add each individual file as a potential test case
        testcases.extend(files)
        return testcases



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import sys  # mypy looks for `sys.version_info`
from contextlib import suppress
from importlib.metadata import EntryPoint, PackageNotFoundError, entry_points, version
from logging import getLogger
from os import getenv
from pathlib import Path
from tempfile import gettempdir
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator

__all__ = (
    "HARNESS_FILE",
    "grz_tmp",
    "iter_entry_points",
    "package_version",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


GRZ_TMP = Path(getenv("GRZ_TMP", gettempdir()), "grizzly")
HARNESS_FILE = Path(__file__).parent / "harness.html"
LOG = getLogger(__name__)


def grz_tmp(*subdir: str | Path) -> Path:
    """Create (if needed) a temporary working directory in a known location.

    Args:
        subdir: Nested directories.

    Returns:
        Path within the temporary working directory.
    """
    path = Path(GRZ_TMP, *subdir)
    path.mkdir(parents=True, exist_ok=True)
    return path


def iter_entry_points(group: str) -> Generator[EntryPoint]:  # pragma: no cover
    """Compatibility wrapper code for importlib.metadata.entry_points()

    Args:
        group: See entry_points().

    Yields:
        EntryPoint
    """
    # TODO: remove this function when support for Python 3.9 is dropped
    assert group
    if sys.version_info >= (3, 10):
        yield from entry_points().select(group=group)
    else:
        raise AssertionError("Unsupported Python version")


def package_version(name: str, default: str = "unknown") -> str:
    """Get version of an installed package.

    Args:
        name: Package name.
        default: String to use if package is not found.

    Returns:
        Version string.
    """
    with suppress(PackageNotFoundError):
        return version(name)
    # package is not installed
    return default



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from hashlib import sha1
from logging import getLogger
from os import SEEK_END
from pathlib import Path
from platform import machine, system
from re import DOTALL, VERBOSE
from re import compile as re_compile
from shutil import copyfileobj, move, rmtree
from tempfile import mkstemp
from time import strftime
from typing import NamedTuple, cast

# import FuzzManager utilities
from Collector.Collector import Collector
from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo, CrashSignature

from .stack_hasher import IGNORED_FRAMES, Stack
from .utils import grz_tmp

__all__ = ("Report",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


# NOTE: order matters, aux -> stderr -> stdout
class LogMap(NamedTuple):
    aux: Path | None
    stderr: Path | None
    stdout: Path | None


class Report:
    DEFAULT_MAJOR = "NO_STACK"
    DEFAULT_MINOR = "NO_STACK"
    HANG_STACK_HEIGHT = 10
    MAX_LOG_SIZE = 1_048_576  # 1MB

    __slots__ = (
        "_crash_info",
        "_logs",
        "_short_signature",
        "_signature",
        "_target_binary",
        "is_hang",
        "path",
        "prefix",
        "stack",
        "unstable",
    )

    def __init__(
        self,
        log_path: Path,
        target_binary: Path,
        is_hang: bool = False,
        size_limit: int = MAX_LOG_SIZE,
        unstable: bool = False,
    ) -> None:
        self._crash_info: CrashInfo | None = None
        self._logs = self._select_logs(log_path)
        assert self._logs is not None
        self._short_signature: str | None = None
        self._signature = None
        self._target_binary = target_binary
        self.is_hang = is_hang
        self.path = log_path
        # if a build crashes before the initial testcase is served it is unstable
        self.unstable = unstable
        # tail files in log_path if needed
        if size_limit < 1:
            LOG.warning("No limit set on report log size!")
        else:
            for log in log_path.iterdir():
                if log.is_file() and log.stat().st_size > size_limit:
                    Report.tail(log, size_limit)
        # look through logs one by one until we find a stack
        for log_file in (x for x in self._logs if x is not None):
            stack = Stack.from_text(log_file.read_text("utf-8", errors="ignore"))
            if stack.frames:
                assert stack.minor is not None
                # limit the hash calculations to the first n frames if a hang
                # was detected to attempt to help local bucketing
                stack.height_limit = self.HANG_STACK_HEIGHT if is_hang else 0
                self.prefix = f"{stack.minor[:8]}_{strftime('%Y-%m-%d_%H-%M-%S')}"
                self.stack: Stack | None = stack
                break
        else:
            self.prefix = f"{self.DEFAULT_MINOR}_{strftime('%Y-%m-%d_%H-%M-%S')}"
            self.stack = None

    @staticmethod
    def calc_hash(signature: CrashSignature | None) -> str:
        """Create unique hash from a signature.

        Args:
            signature:

        Returns:
            Hash of the raw signature.
        """
        if signature is None:
            return "NO_SIGNATURE"
        return sha1(signature.rawSignature.encode("utf-8")).hexdigest()[:16]

    def cleanup(self) -> None:
        """Remove Report data from filesystem.

        Args:
            None

        Returns:
            None
        """
        if self.path.is_dir():
            rmtree(self.path)

    @property
    def crash_hash(self) -> str:
        """Create unique hash from a signature.

        Args:
            None

        Returns:
            Hash of the raw signature of the crash.
        """
        if self.is_hang:
            # TODO: we cannot create a unique bucket hash for hangs atm
            return "hang"
        return self.calc_hash(self.crash_signature)

    @property
    def crash_info(self) -> CrashInfo:
        """Create CrashInfo object from logs.

        Args:
            None

        Returns:
            CrashInfo based on log data.
        """
        if self._crash_info is None:
            assert self._logs is not None
            assert self.path is not None
            # create ProgramConfiguration that can be reported to a FM server
            if Path(f"{self._target_binary}.fuzzmanagerconf").is_file():
                # attempt to use "<target_binary>.fuzzmanagerconf"
                fm_cfg = ProgramConfiguration.fromBinary(self._target_binary)
            else:
                LOG.debug("'%s.fuzzmanagerconf' does not exist", self._target_binary)
                LOG.debug("creating ProgramConfiguration")
                cpu = machine().lower()
                fm_cfg = ProgramConfiguration(
                    self._target_binary.name,
                    "x86_64" if cpu == "amd64" else cpu,
                    system(),
                )
            # read the log files and create a CrashInfo object
            self._crash_info = CrashInfo.fromRawCrashData(
                self._load_log(self._logs.stdout) if self._logs.stdout else None,
                self._load_log(self._logs.stderr) if self._logs.stderr else None,
                fm_cfg,
                auxCrashData=self._load_log(self._logs.aux) if self._logs.aux else None,
            )
        return self._crash_info

    @property
    def crash_signature(self) -> CrashSignature | None:
        """Create CrashSignature object from CrashInfo.

        Args:
            None

        Returns:
            CrashSignature based on log data.
        """
        if self._signature is None:
            collector = Collector()
            if collector.sigCacheDir and Path(collector.sigCacheDir).is_dir():
                cache_sig, _ = collector.search(self.crash_info)
                if cache_sig:
                    LOG.debug("signature loaded from cache file %r", cache_sig)
                    self._signature = CrashSignature.fromFile(cache_sig)
            # if cache lookup failed generate a crash signature
            if self._signature is None:
                self._signature = self.crash_info.createCrashSignature(
                    maxFrames=self.crash_signature_max_frames(self.crash_info)
                )
            if self._signature is None:
                LOG.debug("failed to create FM signature")
        return self._signature

    @staticmethod
    def crash_signature_max_frames(
        crash_info: CrashInfo, suggested_frames: int = 8
    ) -> int:
        """Determine how many stack frames should be used when creating a signature.

        Args:
            crash_info: Data to analyse.
            suggested_frames: Minimum number of frames to use when creating a signature.

        Returns:
            Number of frames to use when creating a signature.
        """
        ignore = 0
        for count, entry in enumerate(crash_info.backtrace, start=1):
            # Rust panics add frames of noise to the top of the stack (std::panicking)
            # Sanitizer heap profile also have more noise on the stack (alloc::alloc)
            if any(entry.startswith(x) for x in IGNORED_FRAMES) or entry.startswith(
                "alloc::alloc"
            ):
                ignore += 1
            # only look at the top of the stack
            if count - ignore == suggested_frames:
                break
        return suggested_frames + ignore

    @staticmethod
    def _find_ffpuppet_worker(logs: list[Path]) -> Path | None:
        """Search through list of log files for a ffpuppet worker log.

        Args:
            logs: List of log files to search.

        Returns:
            Log file if a match is found otherwise None.
        """
        found = None
        for log_file in (x for x in logs if "ffp_worker" in x.name):
            if found is not None:
                # we only expect one log here...
                LOG.warning("overwriting previously selected '%s'", log_file)
            found = log_file
        return found

    @staticmethod
    def _find_minidump(logs: list[Path]) -> Path | None:
        """Search through list of log files for a minidump log.

        Args:
            logs: List of log files to search.

        Returns:
            Log file if a match is found otherwise None.
        """
        re_dump_req = re_compile(
            r"\d+\|0\|.+?\|google_breakpad::ExceptionHandler::WriteMinidump"
        )
        for log_file in (x for x in logs if "minidump" in x.name):
            with log_file.open() as log_fp:
                data = log_fp.read(65536)
                # this will select log that contains "Crash|SIGSEGV|" or
                # the desired "DUMP_REQUESTED" log
                # TODO: review this it may be too strict
                # see mozilla-central/source/accessible/ipc/DocAccessibleParent.cpp#452
                if "Crash|DUMP_REQUESTED|" not in data or re_dump_req.search(data):
                    return log_file
        return None

    @staticmethod
    def _find_sanitizer(logs: list[Path]) -> Path | None:
        """Search through list of log files for a sanitizer (ASan, UBSan, etc...) log.

        Args:
            logs: List of log files to search.

        Returns:
            Log file if a match is found otherwise None.
        """
        # pattern to identify the ASan crash triggered when the parent process goes away
        # TODO: this may no longer be required
        re_e10s_forced = re_compile(
            r"""
            ==\d+==ERROR:.+?SEGV\son.+?0x[0]+\s\(.+?T2\).+?
            #0\s+0x[0-9a-f]+\sin\s+mozilla::ipc::MessageChannel::OnChannelErrorFromLink
            """,
            DOTALL | VERBOSE,
        )
        # this is a list of Sanitizer error reports to prioritize
        # Sanitizer reports not included below are deprioritized
        prioritize_tokens = (
            "use-after-",
            "-buffer-overflow on",
            ": data race ",
            ": SEGV on ",
            "access-violation on ",
            "attempting free on ",
            "negative-size-param",
            "-param-overlap",
        )
        fallback = None
        found = None
        for log_file in (x for x in logs if "asan" in x.name):
            with log_file.open() as log_fp:
                data = log_fp.read(65536)
            # look for interesting crash info in the log
            if "==ERROR:" in data or "WARNING:" in data:
                # check for e10s forced crash
                if re_e10s_forced.search(data) is not None:
                    continue
                # make sure there is something that looks like a stack frame
                if "#0 " in data:
                    found = log_file
                    if any(x in data for x in prioritize_tokens):
                        # this is the likely cause of the crash
                        break
            if found is None:
                # UBSan error (non-ASan builds)
                if ": runtime error: " in data:
                    found = log_file
                # catch all (choose the one with info for now)
                elif data:
                    fallback = log_file
        return found or fallback

    @staticmethod
    def _find_valgrind(logs: list[Path]) -> Path | None:
        """Search through log files for a Valgrind log. Empty files are ignored.

        Args:
            logs: List of log files to search.

        Returns:
            Log file if a match is found otherwise None.
        """
        for log_file in logs:
            if "valgrind" in log_file.name and log_file.stat().st_size:
                return log_file
        return None

    @staticmethod
    def _load_log(path: Path) -> list[str]:
        """Load and sanitize text from a file for use with CrashInfo.fromRawCrashData().

        Args:
            path: Log file to load.

        Returns:
            Text data sanitized and split into lines.
        """
        return path.read_text("utf-8", errors="replace").replace("\0", "?").splitlines()

    @property
    def major(self) -> str:
        """The inclusive bucketing hash based on the stack trace data found in logs.

        Args:
            None

        Returns:
            Major hash string.
        """
        if self.stack and self.stack.major is not None:
            return self.stack.major
        return self.DEFAULT_MAJOR

    @property
    def minor(self) -> str:
        """The specific bucketing hash based on the stack trace data found in logs.

        Args:
            None

        Returns:
            Minor hash string.
        """
        if self.stack and self.stack.minor is not None:
            return self.stack.minor
        return self.DEFAULT_MINOR

    @property
    def preferred(self) -> Path | None:
        """Log file containing most relevant data.

        Args:
            None

        Returns:
            Log file.
        """
        assert self._logs is not None
        return self._logs.aux or self._logs.stderr

    @classmethod
    def _select_logs(cls, path: Path) -> LogMap | None:
        """Scan path for file containing stderr, stdout and other (aux)
        data and build a LogMap.

        Args:
            path: Location to scan for log files.

        Returns:
            A LogMap pointing to log files or None if path is empty.
        """
        files = (x for x in path.iterdir() if x.is_file())
        # order by date hopefully the oldest log is the cause of the issue
        to_scan = sorted(files, key=lambda x: x.stat().st_mtime)
        if not to_scan:
            LOG.warning("No files found in '%s'", path)
            return None
        # look for file to use as aux log
        log_aux = cls._find_sanitizer(to_scan)
        if log_aux is None:
            log_aux = cls._find_valgrind(to_scan)
        if log_aux is None:
            log_aux = cls._find_minidump(to_scan)
        if log_aux is None:
            log_aux = cls._find_ffpuppet_worker(to_scan)
        # look for stderr and stdout log files
        log_err = None
        log_out = None
        for log_files in to_scan:
            if "stderr" in log_files.name:
                log_err = log_files
            elif "stdout" in log_files.name:
                log_out = log_files
        result = LogMap(log_aux, log_err, log_out)
        return result if any(result) else None

    @property
    def short_signature(self) -> str:
        """Short signature of the report created by FuzzManager.

        Args:
            None

        Returns:
            Short signature.
        """
        if self._short_signature is None:
            if self.is_hang:
                # TODO: remove once we can create accurate signatures for hangs
                self._short_signature = "Potential hang detected"
            elif self.crash_signature is None:
                # FM crash signature creation failed
                self._short_signature = "Signature creation failed"
            else:
                # TODO: FM is missing type hints
                self._short_signature = cast(
                    str, self.crash_info.createShortSignature()
                )
        return self._short_signature

    @staticmethod
    def tail(in_file: Path, size_limit: int) -> None:
        """Tail the given file. WARNING: This is destructive!

        Args:
            in_file: File to work with.
            size_limit: Maximum size of file after tail operation.

        Returns:
            None
        """
        assert size_limit > 0
        with in_file.open("rb") as in_fp:
            in_fp.seek(0, SEEK_END)
            end = in_fp.tell()
            if end <= size_limit:
                return
            dump_pos = end - size_limit
            in_fp.seek(dump_pos)
            out_fd, out_file = mkstemp(prefix="taillog_", dir=grz_tmp())
            with open(out_fd, "wb") as out_fp:
                out_fp.write(b"[LOG TAILED]\n")
                copyfileobj(in_fp, out_fp, 0x10000)  # 64KB chunks
        in_file.unlink()
        move(out_file, in_file)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from contextlib import suppress
from logging import getLogger
from shutil import move, rmtree
from time import time
from typing import TYPE_CHECKING

from fasteners import InterProcessLock

from .utils import grz_tmp

if TYPE_CHECKING:
    from pathlib import Path


_ACTIVE_CACHE = None
CACHE_PATH = grz_tmp("cache")
CACHE_TIME = int(time())
LOCK_FILE = grz_tmp() / "cache.lock"
LOG = getLogger(__name__)
MAX_AGE = 86400


def _active_cache(max_age: int = MAX_AGE) -> Path:
    """Retrieve the active cache directory. Create one if needed.

    Args:
        max_age: Maximum age of active cache (relative to process launch time).

    Returns:
        Directory to use to store and retrieve cached content.
    """
    global _ACTIVE_CACHE  # pylint: disable=global-statement
    assert max_age >= 0
    if _ACTIVE_CACHE is None:
        limit = CACHE_TIME - max_age
        with InterProcessLock(LOCK_FILE):
            # find most recent active entry
            # TODO: ideally this should use the creation time not the directory name
            # but that is not currently available on all platforms
            for entry in sorted(
                (
                    int(x.name)
                    for x in CACHE_PATH.iterdir()
                    if x.is_dir() and x.name.isdecimal()
                ),
                reverse=True,
            ):
                if entry > limit:
                    _ACTIVE_CACHE = CACHE_PATH / str(entry)
                    LOG.debug("active cache found: '%s'", _ACTIVE_CACHE)
                    break
            else:
                # create a new active entry if one does not exist
                _ACTIVE_CACHE = CACHE_PATH / str(CACHE_TIME)
                _ACTIVE_CACHE.mkdir(parents=True)
                LOG.debug("active cache created: '%s'", _ACTIVE_CACHE)
    return _ACTIVE_CACHE


def _valid_key(key: str) -> bool:
    """Check if key contains only alphanumeric characters. Dash/hyphen (-) is allowed.

    Args:
        key: Identifier to validate.

    Returns:
        True if key is valid otherwise False.
    """
    return key.replace("-", "").isalnum()


def add_cached(key: str, src: Path) -> Path:
    """Move a file or directory into the cache.

    Args:
        key: Identifier used to lookup cached data.
        src: File or directory to cache.

    Returns:
        Directory containing cached content.
    """
    if not _valid_key(key):
        raise ValueError("Key must be alphanumeric")
    dst = _active_cache() / key
    with InterProcessLock(LOCK_FILE):
        dst.mkdir(parents=True, exist_ok=True)
        if (dst / src.name).exists():
            LOG.debug("add_cache: '%s' exists in '%s'", (dst / src.name), dst)
        else:
            move(src, dst)
    return dst


def clear_cached(max_age: int = MAX_AGE * 2) -> None:
    """Remove expired content from cache.

    Args:
        max_age: Maximum age of cache directory (relative to process launch time).

    Returns:
        None
    """
    assert max_age >= 0
    limit = CACHE_TIME - max_age
    with InterProcessLock(LOCK_FILE):
        # iterate over all directories in CACHE_PATH
        for entry in (x for x in CACHE_PATH.iterdir() if x.is_dir()):
            with suppress(ValueError):
                # remove only expired
                if int(entry.name) <= limit:
                    LOG.debug("removing old cache entry: '%s'", entry)
                    rmtree(entry, ignore_errors=True)


def find_cached(key: str) -> Path | None:
    """Find data in local cache.

    Args:
        key: Identifier used to lookup cached data.

    Returns:
        Directory containing cached content or None if no valid entry is found.
    """
    if not _valid_key(key):
        raise ValueError("Key must be alphanumeric")
    path = _active_cache() / key
    with InterProcessLock(LOCK_FILE):
        if path.is_dir():
            return path
    return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer testcase list strategy definition."""
from __future__ import annotations

from logging import getLogger
from shutil import rmtree
from typing import TYPE_CHECKING

from ...common.storage import TestCase
from . import Strategy

if TYPE_CHECKING:
    from collections.abc import Generator

LOG = getLogger(__name__)


class MinimizeTestcaseList(Strategy):
    """Try removing testcases from a list of sequential testcases (eg. Grizzly result
    cache). The strategy favours testcases at the tail of the list, so for a list of
    five testcases:
                            testcases
                    0       1 2 3 4 5
        iteration   1         2 3 4 5
                    2       1   3 4 5
                    3       1 2   4 5
                    4       1 2 3   5
                    5       1 2 3 4
    """

    name = "list"

    def __init__(self, testcases: list[TestCase]) -> None:
        """Initialize strategy instance.

        Arguments:
            testcases: Testcases to reduce. The object does not take ownership of the
                       testcases.
        """
        super().__init__(testcases)
        self._current_feedback: bool | None = None
        # TODO: is this unused?
        self._current_served = None

    def update(self, success: bool) -> None:
        """Inform the strategy whether or not the last reduction yielded was good.

        Arguments:
            success: Whether or not the last reduction was acceptable.

        Returns:
            None
        """
        assert self._current_feedback is None
        assert self._current_served is None
        self._current_feedback = success

    def __iter__(self) -> Generator[list[TestCase]]:
        """Iterate over potential reductions of testcases according to this strategy.

        The caller should evaluate each testcase set yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            Testcases with reduction applied
        """
        assert self._current_feedback is None
        idx = 0
        testcases = [TestCase.load(x) for x in sorted(self._testcase_root.iterdir())]
        n_testcases = len(testcases)
        while True:
            if n_testcases <= 1:
                LOG.info(
                    "Testcase list has length %d, not enough to reduce!",
                    n_testcases,
                )
                break
            if idx >= n_testcases:
                LOG.info("Attempted to remove every single testcase")
                break
            # try removing the testcase at idx
            excluded = testcases.pop(idx)
            yield testcases
            assert self._current_feedback is not None, "no feedback received!"
            if self._current_feedback:
                rmtree(excluded.root, ignore_errors=True)
                LOG.info(
                    "Removing testcase %d/%d was successful!", idx + 1, n_testcases
                )
                n_testcases = len(testcases)
            else:
                testcases.insert(idx, excluded)
                LOG.info("No result without testcase %d/%d", idx + 1, n_testcases)
                idx += 1
            # reset
            self._current_feedback = None
            self._current_served = None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from time import perf_counter
from traceback import format_exception
from typing import TYPE_CHECKING, Callable

from .worker import Worker

if TYPE_CHECKING:
    from socket import socket

    from .job import Job

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class ConnectionManager:
    # allow extra time before closing socket if needed
    SHUTDOWN_DELAY = 0.5

    __slots__ = (
        "_deadline",
        "_deadline_exceeded",
        "_job",
        "_limit",
        "_next_poll",
        "_poll",
        "_socket",
    )

    def __init__(
        self, job: Job, srv_socket: socket, limit: int = 1, poll: float = 0.5
    ) -> None:
        assert limit > 0
        assert poll > 0
        self._deadline: float | None = None
        self._deadline_exceeded = False
        self._job = job
        self._limit = limit
        self._next_poll = 0.0
        self._poll = poll
        self._socket = srv_socket

    def __enter__(self) -> ConnectionManager:
        return self

    def __exit__(self, *exc: object) -> None:
        self.close()

    def _can_continue(self, continue_cb: Callable[[], bool] | None) -> bool:
        """Check timeout and callback status.

        Args:
            continue_cb: Indicates whether to continue.

        Returns:
            True if callback returns True and timeout has been be hit otherwise False.
        """
        now = perf_counter()
        if self._next_poll > now:
            return True
        self._next_poll = now + self._poll
        # check if callback returns False
        if continue_cb is not None and not continue_cb():
            LOG.debug("continue_cb() returned False")
            return False
        # check for a timeout
        if self._deadline and self._deadline <= now:
            LOG.debug("exceeded serve deadline")
            self._deadline_exceeded = True
            return False
        return True

    def close(self) -> None:
        """Set job state to finished and raise any errors encountered by workers.

        Args:
            None

        Returns:
            None
        """
        self._job.finish()
        if not self._job.exceptions.empty():
            exc_type, exc_obj, exc_tb = self._job.exceptions.get()
            LOG.error(
                "Unexpected exception:\n%s",
                "".join(format_exception(exc_type, exc_obj, exc_tb)),
            )
            # re-raise exception from worker once all workers are closed
            raise exc_obj

    @staticmethod
    def _join_workers(workers: list[Worker], timeout: float = 0) -> list[Worker]:
        """Attempt to join workers.

        Args:
            workers: Collection of workers.
            timeout: Maximum time in seconds to wait.

        Returns:
            Workers that do not join before the timeout is reached.
        """
        deadline = perf_counter() + timeout
        return [
            x for x in workers if not x.join(timeout=max(deadline - perf_counter(), 0))
        ]

    def serve(
        self,
        timeout: int,
        continue_cb: Callable[[], bool] | None = None,
        shutdown_delay: float = SHUTDOWN_DELAY,
    ) -> bool:
        """Manage workers and serve job contents.

        Args:
            timeout: Maximum time to serve in seconds.
            continue_cb: Indicates whether to continue.
            shutdown_delay: Time in seconds to wait before calling shutdown on
                            sockets of active workers.

        Returns:
            True unless the timeout is exceeded.
        """
        assert self._job.pending or self._job.forever
        assert shutdown_delay >= 0
        assert timeout >= 0
        if continue_cb is not None and not callable(continue_cb):
            raise TypeError("continue_cb must be callable")

        self._deadline_exceeded = False
        start_time = perf_counter()
        self._deadline = start_time + timeout if timeout else None

        launches = 0
        running = 0
        workers: list[Worker] = []
        LOG.debug("accepting requests (workers: %d, timeout: %r)", self._limit, timeout)
        try:
            while not self._job.is_complete() and self._can_continue(continue_cb):
                # launch workers
                if running < self._limit:
                    if not self._job.accepting.wait(0.05):
                        # wait for accepting flag to be set
                        continue
                    worker = Worker.launch(self._socket, self._job)
                    if worker is not None:
                        workers.append(worker)
                        running = len(workers)
                        launches += 1

                # manage workers
                if running >= self._limit:
                    LOG.debug("worker limit (%d) hit, waiting...", running)
                    if self._job.worker_complete.wait(1):
                        self._job.worker_complete.clear()
                    workers = self._join_workers(workers)
                    running = len(workers)
                    LOG.debug("removed completed workers (%d active)", running)

        finally:
            LOG.debug(
                "serve exit: %d request(s) in %0.3fs, waiting for %d worker(s)...",
                launches,
                perf_counter() - start_time,
                running,
            )
            if not self._job.is_complete():
                LOG.debug("job not complete")
                self._job.finish()
            # use shutdown_delay to avoid cutting off connections
            workers = self._join_workers(workers, timeout=shutdown_delay)
            # close remaining active workers
            if workers:
                LOG.debug("closing remaining active workers: %d", len(workers))
                for worker in workers:
                    worker.close()
                # join remaining workers
                if self._join_workers(workers, timeout=30):
                    LOG.error("Failed to close %d workers", len(workers))
                    raise RuntimeError("Failed to close workers")

        # return False only if there was a timeout
        return not self._deadline_exceeded



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from hashlib import sha1
from logging import getLogger
from os import getpid, kill
from pathlib import Path
from platform import system
from shutil import copytree, rmtree
from signal import SIGABRT
from tempfile import TemporaryDirectory, mkdtemp
from typing import TYPE_CHECKING, Any

from ffpuppet import BrowserTimeoutError, Debugger, FFPuppet, LaunchError, Reason
from ffpuppet.display import DisplayMode
from ffpuppet.helpers import certutil_available, certutil_find
from ffpuppet.profile import Profile
from ffpuppet.sanitizer_util import SanitizerOptions
from prefpicker import PrefPicker

from ..common.cache import add_cached, find_cached
from ..common.report import Report
from ..common.utils import grz_tmp, package_version
from .target import Result, Target, TargetLaunchError, TargetLaunchTimeout
from .target_monitor import TargetMonitor

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from sapphire import CertificateBundle

__all__ = ("PuppetTarget",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger(__name__)


class PuppetMonitor(TargetMonitor):
    def __init__(self, puppet: FFPuppet) -> None:
        self._puppet = puppet

    def clone_log(self, log_id: str, offset: int = 0) -> Path | None:
        return self._puppet.clone_log(log_id, offset=offset)

    def is_healthy(self) -> bool:
        return self._puppet.is_healthy()

    def is_idle(self, threshold: int) -> bool:
        # assert 0 <= threshold <= 100
        return all(cpu < threshold for _, cpu in self._puppet.cpu_usage())

    def is_running(self) -> bool:
        return self._puppet.is_running()

    @property
    def launches(self) -> int:
        return self._puppet.launches

    def log_length(self, log_id: str) -> int:
        return self._puppet.log_length(log_id) or 0


class PuppetTarget(Target):
    SUPPORTED_ASSETS = (
        # file containing line separated list of tokens to scan stderr/out for
        "abort-tokens",
        # xpi or directory containing the unpacked extension
        "extension",
        # LSan suppression list file
        "lsan-suppressions",
        # prefs.js file to use
        "prefs",
        # TSan suppression list file
        "tsan-suppressions",
        # UBSan suppression list file
        "ubsan-suppressions",
    )

    TRACKED_ENVVARS = (
        "ASAN_OPTIONS",
        "LSAN_OPTIONS",
        "TSAN_OPTIONS",
        "UBSAN_OPTIONS",
        "GNOME_ACCESSIBILITY",
        "MOZ_CHAOSMODE",
        "MOZ_FUZZ_CRASH_ON_LARGE_ALLOC",
        "MOZ_FUZZ_HAS_GPU",
        "MOZ_FUZZ_LARGE_ALLOC_LIMIT",
        "XPCOM_DEBUG_BREAK",
    )

    __slots__ = (
        "_debugger",
        "_extension",
        "_monitor",
        "_prefs",
        "_profile_template",
        "_puppet",
        "use_valgrind",
    )

    def __init__(
        self,
        binary: Path,
        launch_timeout: int,
        log_limit: int,
        memory_limit: int,
        certs: CertificateBundle | None = None,
        display_mode: str = "default",
        pernosco: bool = False,
        rr: bool = False,
        valgrind: bool = False,
        **kwds: dict[str, Any],
    ) -> None:
        super().__init__(
            binary,
            launch_timeout,
            log_limit,
            memory_limit,
        )
        LOG.debug("ffpuppet version: %s", package_version("ffpuppet"))
        # only pass certs to FFPuppet if certutil is available
        # otherwise certs can't be used
        self._https = False
        self._monitor: PuppetMonitor | None = None
        self._profile_template = None
        if certs is not None:
            certutil = certutil_find(binary)
            if certutil_available(certutil):
                self._https = True
                self._profile_template = self._get_certdb(certs.root, certutil)
            else:
                LOG.warning("HTTPS support requires NSS certutil.")

        # TODO: clean up handling debuggers
        self._debugger = Debugger.NONE
        if pernosco:
            self._debugger = Debugger.PERNOSCO
        if rr:
            self._debugger = Debugger.RR
        if valgrind:
            # TODO: replace use_valgrind with something debugger generic
            self.use_valgrind = True
            self._debugger = Debugger.VALGRIND
        self._extension: Path | None = None
        self._prefs: Path | None = None

        # create FFPuppet object
        self._puppet = FFPuppet(
            debugger=self._debugger,
            display_mode=DisplayMode[display_mode.upper()],
            use_profile=self._profile_template,
            working_path=str(grz_tmp("target")),
        )
        if kwds:
            LOG.debug("PuppetTarget ignoring unsupported kwargs: %s", ", ".join(kwds))

    def _cleanup(self) -> None:
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.clean_up()
        if self._profile_template is not None:
            rmtree(self._profile_template)

    @staticmethod
    def _get_certdb(cert_file: Path, certuil: str) -> Path:
        """Load or create a certdb to use with Firefox. This will search for a cached
        certificate and certdb before generating and caching a new bundle.
        The cache allows reuse across processes.

        Args:
            cert_file: Certificate to install into the newly created certdb.
            certutil: certutil binary.

        Returns:
            Directory containing the certdb files.
        """
        cached = find_cached("crypto")
        # certs and certdb are dependent on each other
        # use the hash of the cert file to ensure this
        db_path = (
            f"certdb_{sha1(cert_file.read_bytes(), usedforsecurity=False).hexdigest()}"
        )
        if cached is None or not (cached / db_path).exists():
            with TemporaryDirectory() as tmp_path:
                certdb = Path(tmp_path) / db_path
                certdb.mkdir(parents=True)
                Profile.init_cert_db(certdb, certuil)
                Profile.install_cert(certdb, cert_file, certuil)
                # add certdb to cache
                cached = add_cached("crypto", certdb)
        # copy data from cache
        path = grz_tmp("certdb") / str(getpid())
        copytree(cached / db_path, path)
        return path

    def close(self, force_close: bool = False) -> None:
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.close(force_close=force_close)

    @property
    def closed(self) -> bool:
        return self._puppet.reason is not None

    def create_report(self, is_hang: bool = False, unstable: bool = False) -> Report:
        logs = Path(mkdtemp(prefix="logs_", dir=grz_tmp("logs")))
        self.save_logs(logs)
        return Report(logs, self.binary, is_hang=is_hang, unstable=unstable)

    def filtered_environ(self) -> dict[str, str]:
        # remove context specific entries from environment
        filtered = dict(self.environ)
        opts = SanitizerOptions()
        # iterate over *SAN_OPTIONS entries
        for san in (x for x in filtered if x.endswith("SAN_OPTIONS")):
            opts.load_options(filtered[san])
            # remove entries specific to the current environment
            opts.pop("external_symbolizer_path")
            opts.pop("log_path")
            opts.pop("strip_path_prefix")
            opts.pop("suppressions")
            filtered[san] = str(opts)
        # remove empty entries
        return {k: v for k, v in filtered.items() if v}

    @property
    def monitor(self) -> PuppetMonitor:
        if self._monitor is None:
            self._monitor = PuppetMonitor(self._puppet)
        return self._monitor

    def check_result(self, ignored: Iterable[str]) -> Result:
        result = Result.NONE
        # check if there has been a crash, hangs will appear as SIGABRT
        if not self._puppet.is_healthy():
            self.close()
            # something has happened figure out what
            if self._puppet.reason == Reason.CLOSED:
                LOG.debug("target.close() was called")
            elif self._puppet.reason == Reason.EXITED:
                LOG.debug("target closed itself")
            elif (
                self._puppet.reason == Reason.WORKER
                and "memory" in ignored
                and "ffp_worker_memory_usage" in self._puppet.available_logs()
            ):
                result = Result.IGNORED
                LOG.debug("memory limit exceeded")
            elif (
                self._puppet.reason == Reason.WORKER
                and "log-limit" in ignored
                and "ffp_worker_log_size" in self._puppet.available_logs()
            ):
                result = Result.IGNORED
                LOG.debug("log size limit exceeded")
            else:
                assert self._puppet.reason is not None
                # crash or hang (forced SIGABRT) has been detected
                LOG.debug("result detected (%s)", self._puppet.reason.name)
                result = Result.FOUND
        return result

    def handle_hang(
        self, ignore_idle: bool = True, ignore_timeout: bool = False
    ) -> bool:
        # only send SIGABRT in certain case
        send_abort = (
            not ignore_timeout
            and system() == "Linux"
            and self._debugger == Debugger.NONE
        )
        was_idle = False
        if self._puppet.is_healthy():
            proc_usage = sorted(self._puppet.cpu_usage(), key=lambda x: x[1])
            if proc_usage:
                pid, cpu = proc_usage.pop()
                if ignore_idle and cpu < 15:
                    # don't send SIGABRT if process is idle
                    LOG.debug("ignoring idle hang (%0.1f%%)", cpu)
                    was_idle = True
                elif send_abort:
                    # sending SIGABRT is only supported on Linux for now
                    # TODO: add/test on other OSs
                    LOG.debug("sending SIGABRT to %r (%0.1f%%)", pid, cpu)
                    try:
                        kill(pid, SIGABRT)
                    except OSError:
                        LOG.warning("Failed to send SIGABRT to pid %d", pid)
                    self._puppet.wait(timeout=10)
        # always call close() since this function should only/always
        # be called when there has been a timeout
        self.close()
        return was_idle

    def https(self) -> bool:
        return self._https

    def dump_coverage(self, timeout: int = 15) -> None:
        if self._puppet.is_healthy():
            self._puppet.dump_coverage(timeout=timeout)

    def launch(self, location: str) -> None:
        # setup environment
        env_mod: dict[str, str | None] = dict(self.environ)
        # do not allow network connections to non local endpoints
        env_mod["MOZ_DISABLE_NONLOCAL_CONNECTIONS"] = "1"
        # we always want the browser to exit when a crash is detected
        env_mod["MOZ_CRASHREPORTER_SHUTDOWN"] = "1"
        try:
            self._puppet.launch(
                self.binary,
                launch_timeout=self.launch_timeout,
                location=location,
                log_limit=self.log_limit,
                memory_limit=self.memory_limit,
                prefs_js=self._prefs,
                extension=[self._extension] if self._extension else None,
                env_mod=env_mod,
            )
        except LaunchError as exc:
            self.close()
            if isinstance(exc, BrowserTimeoutError):
                raise TargetLaunchTimeout(str(exc)) from None
            raise TargetLaunchError(
                str(exc), self.create_report(unstable=True)
            ) from None

    def log_size(self) -> int:
        total = 0
        for log in ("stderr", "stdout"):
            length = self._puppet.log_length(log)
            if length:
                total += length
        return total

    def merge_environment(self, extra: Mapping[str, str]) -> None:
        output = dict(extra)
        if self.environ:
            # prioritize existing environment variables
            output.update(self.environ)
            # merge contents of *SAN_OPTIONS
            org = SanitizerOptions()
            out = SanitizerOptions()
            for san in ("ASAN", "LSAN", "TSAN", "UBSAN"):
                opts = f"{san}_OPTIONS"
                org.load_options(self.environ.get(opts, ""))
                if not org:
                    # nothing to add from original
                    continue
                out.load_options(extra.get(opts, ""))
                for opt, value in org:  # pylint: disable=not-an-iterable
                    out.add(opt, value, overwrite=True)
                output[opts] = str(out)
        self.environ = output

    def process_assets(self) -> None:
        self._extension = self.asset_mgr.get("extension")
        self._prefs = self.asset_mgr.get("prefs")
        # generate temporary prefs.js with prefpicker
        if self._prefs is None:
            LOG.debug("using prefpicker to generate prefs.js")
            with TemporaryDirectory(dir=grz_tmp("target")) as tmp_path:
                prefs = Path(tmp_path) / "prefs.js"
                template = PrefPicker.lookup_template("browser-fuzzing.yml")
                assert template
                PrefPicker.load_template(template).create_prefsjs(prefs)
                self._prefs = self.asset_mgr.add("prefs", prefs, copy=False)
        abort_tokens = self.asset_mgr.get("abort-tokens")
        if abort_tokens:
            LOG.debug("loading 'abort tokens' from '%s'", abort_tokens)
            with (self.asset_mgr.path / abort_tokens).open() as in_fp:
                for line in in_fp:
                    line = line.strip()
                    if line:
                        self._puppet.add_abort_token(line)

        # configure sanitizer suppressions
        opts = SanitizerOptions()
        for sanitizer in ("lsan", "tsan", "ubsan"):
            asset = f"{sanitizer}-suppressions"
            # load existing sanitizer options from environment
            var_name = f"{sanitizer.upper()}_OPTIONS"
            opts.load_options(self.environ.get(var_name, ""))
            if self.asset_mgr.get(asset):
                # use suppression file if provided as asset
                opts.add(
                    "suppressions", f"'{self.asset_mgr.get(asset)}'", overwrite=True
                )
            elif opts.get("suppressions"):
                suppressions = opts.pop("suppressions")
                assert suppressions
                path = Path(suppressions.strip("\"'"))
                if path.is_file():
                    # use environment specified suppression file
                    LOG.debug("using %r from environment", asset)
                    opts.add(
                        "suppressions",
                        f"'{self.asset_mgr.add(asset, path)}'",
                        overwrite=True,
                    )
                else:
                    LOG.warning("Missing %s suppressions file '%s'", sanitizer, path)
            else:
                LOG.debug("%r does not contain suppressions", var_name)
                continue
            # update sanitized *SAN_OPTIONS
            LOG.debug("updating suppressions in %r", var_name)
            self.environ[var_name] = str(opts)

    def save_logs(self, dst: Path) -> None:
        self._puppet.save_logs(dst)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Common unit test fixtures for `grizzly`.
"""

from pytest import fixture

from .adapter import Adapter
from .target import Target


@fixture
def session_setup(mocker):
    mocker.patch("grizzly.main.FuzzManagerReporter", autospec=True)
    mocker.patch("grizzly.main.Sapphire", autospec_set=True)
    adapter_cls = mocker.Mock(spec_set=Adapter)
    adapter_cls.return_value.RELAUNCH = Adapter.RELAUNCH
    adapter_cls.return_value.TIME_LIMIT = Adapter.TIME_LIMIT
    target_cls = mocker.Mock(spec_set=Target)
    mocker.patch("grizzly.main.load_plugin", side_effect=(adapter_cls, target_cls))
    session_cls = mocker.patch("grizzly.main.Session", autospec_set=True)
    session_obj = session_cls.return_value.__enter__.return_value
    session_obj.status.results.total = 0
    yield adapter_cls.return_value, session_cls, session_obj, target_cls


@fixture
def patch_collector(mocker):
    """Provide a mock Collector to avoid scanning for signatures on disk."""
    collector = mocker.patch("grizzly.common.report.Collector", autospec=True)
    # don't search for signatures locally
    collector.return_value.sigCacheDir = None


@fixture
def tmp_path_status_db_fuzz(tmp_path, mocker):
    """Use a temporary status database file for testing."""
    mocker.patch("grizzly.common.status.STATUS_DB_FUZZ", tmp_path / "fuzzing-tmp.db")
    mocker.patch("grizzly.session.STATUS_DB_FUZZ", tmp_path / "fuzzing-tmp.db")


@fixture
def tmp_path_status_db_reduce(tmp_path, mocker):
    """Use a temporary status database file for testing."""
    mocker.patch("grizzly.common.status.STATUS_DB_REDUCE", tmp_path / "reduce-tmp.db")
    mocker.patch("grizzly.reduce.core.STATUS_DB_REDUCE", tmp_path / "fuzzing-tmp.db")
